import copy
import json
import os
import shutil
from datetime import datetime, timedelta, timezone
from pathlib import Path
from typing import TYPE_CHECKING, Any, Callable, List, Literal, Union, cast

import pandas as pd
import prefect
import pyarrow as pa
import pyarrow.dataset as ds
from prefect import Flow, Task, task
from prefect.backend import set_key_value
from prefect.engine.state import Failed
from prefect.storage import Git
from prefect.tasks.secrets import PrefectSecret
from prefect.utilities import logging
from sendgrid import SendGridAPIClient
from sendgrid.helpers.mail import Mail
from toolz import curry
from visions.functional import infer_type
from visions.typesets.complete_set import CompleteSet

from viadot.config import local_config
from viadot.exceptions import CredentialError
from viadot.tasks import AzureDataLakeUpload, AzureKeyVaultSecret

logger = logging.get_logger()
METADATA_COLUMNS = {"_viadot_downloaded_at_utc": "DATETIME"}


@task(timeout=3600)
def add_ingestion_metadata_task(
    df: pd.DataFrame,
):
    """Add ingestion metadata columns, eg. data download date

    Args:
        df (pd.DataFrame): input DataFrame.
    """
    # Don't skip when df has columns but has no data
    if len(df.columns) == 0:
        return df
    else:
        df2 = df.copy(deep=True)
        df2["_viadot_downloaded_at_utc"] = datetime.now(timezone.utc).replace(
            microsecond=0
        )
        return df2


@task(timeout=3600)
def get_latest_timestamp_file_path(files: List[str]) -> str:
    """
    Return the name of the latest file in a given data lake directory,
    given a list of paths in that directory. Such list can be obtained using the
    `AzureDataLakeList` task. This task is useful for working with immutable data lakes as
    the data is often written in the format /path/table_name/TIMESTAMP.parquet.
    """

    logger = prefect.context.get("logger")

    extract_fname = (
        lambda f: os.path.basename(f).replace(".csv", "").replace(".parquet", "")
    )
    file_names = [extract_fname(file) for file in files]
    latest_file_name = max(file_names, key=lambda d: datetime.fromisoformat(d))
    latest_file = files[file_names.index(latest_file_name)]

    logger.debug(f"Latest file: {latest_file}")

    return latest_file


@task(timeout=3600)
def dtypes_to_json_task(dtypes_dict, local_json_path: str):
    """
    Creates json file from a dictionary.
    Args:
        dtypes_dict (dict): Dictionary containing data types.
        local_json_path (str): Path to local json file.
    """
    with open(local_json_path, "w") as fp:
        json.dump(dtypes_dict, fp)


@task(timeout=3600)
def chunk_df(df: pd.DataFrame, size: int = 10_000) -> List[pd.DataFrame]:
    """
    Creates pandas Dataframes list of chunks with a given size.
    Args:
        df (pd.DataFrame): Input pandas DataFrame.
        size (int, optional): Size of a chunk. Defaults to 10000.
    """
    n_rows = df.shape[0]
    chunks = [df[i : i + size] for i in range(0, n_rows, size)]
    return chunks


@task(timeout=3600)
def df_get_data_types_task(df: pd.DataFrame) -> dict:
    """
    Returns dictionary containing datatypes of pandas DataFrame columns.
    Args:
        df (pd.DataFrame): Input pandas DataFrame.
    """
    typeset = CompleteSet()
    dtypes = infer_type(df, typeset)
    dtypes_dict = {k: str(v) for k, v in dtypes.items()}
    return dtypes_dict


@task(timeout=3600)
def get_sql_dtypes_from_df(df: pd.DataFrame) -> dict:
    """Obtain SQL data types from a pandas DataFrame"""
    typeset = CompleteSet()
    dtypes = infer_type(df.head(10000), typeset)
    dtypes_dict = {k: str(v) for k, v in dtypes.items()}
    dict_mapping = {
        "Float": "REAL",
        "Image": None,
        "Categorical": "VARCHAR(500)",
        "Time": "TIME",
        "Boolean": "VARCHAR(5)",  # Bool is True/False, Microsoft expects 0/1
        "DateTime": "DATETIMEOFFSET",  # DATETIMEOFFSET is the only timezone-aware dtype in TSQL
        "Object": "VARCHAR(500)",
        "EmailAddress": "VARCHAR(50)",
        "File": None,
        "Geometry": "GEOMETRY",
        "Ordinal": "INT",
        "Integer": "INT",
        "Generic": "VARCHAR(500)",
        "UUID": "VARCHAR(50)",  # Microsoft uses a custom UUID format so we can't use it
        "Complex": None,
        "Date": "DATE",
        "String": "VARCHAR(500)",
        "IPAddress": "VARCHAR(39)",
        "Path": "VARCHAR(255)",
        "TimeDelta": "VARCHAR(20)",  # datetime.datetime.timedelta; eg. '1 days 11:00:00'
        "URL": "VARCHAR(255)",
        "Count": "INT",
    }
    dict_dtypes_mapped = {}
    for k in dtypes_dict:
        dict_dtypes_mapped[k] = dict_mapping[dtypes_dict[k]]

    # This is required as pandas cannot handle mixed dtypes in Object columns
    dtypes_dict_fixed = {
        k: ("String" if v == "Object" else str(v))
        for k, v in dict_dtypes_mapped.items()
    }

    return dtypes_dict_fixed


@task(timeout=3600)
def update_dict(d: dict, d_new: dict) -> dict:
    d_copy = copy.deepcopy(d)
    d_copy.update(d_new)
    return d_copy


@task(timeout=3600)
def df_map_mixed_dtypes_for_parquet(
    df: pd.DataFrame, dtypes_dict: dict
) -> pd.DataFrame:
    """
    Pandas is not able to handle mixed dtypes in the column in to_parquet
    Mapping 'object' visions dtype to 'string' dtype to allow Pandas to_parquet

    Args:
        dict_dtypes_mapped (dict): Data types dictionary inferenced by Visions
        df (pd.DataFrame): input DataFrame.

    Returns:
        df_mapped (pd.DataFrame): Pandas DataFrame with mapped Data Types to workaround Pandas to_parquet bug connected with mixed dtypes in object:.
    """
    df_mapped = df.copy()
    for col, dtype in dtypes_dict.items():
        if dtype == "Object":
            df_mapped[col] = df_mapped[col].astype("string")
    return df_mapped


@task(timeout=3600)
def update_dtypes_dict(dtypes_dict: dict) -> dict:
    """
    Task to update dtypes_dictionary that will be stored in the schema. It's required due to workaround Pandas to_parquet bug connected with mixed dtypes in object

    Args:
        dtypes_dict (dict): Data types dictionary inferenced by Visions

    Returns:
        dtypes_dict_updated (dict): Data types dictionary updated to follow Pandas requeirments in to_parquet functionality.
    """
    dtypes_dict_updated = {
        k: ("String" if v == "Object" else str(v)) for k, v in dtypes_dict.items()
    }

    return dtypes_dict_updated


@task(timeout=3600)
def df_to_csv(
    df: pd.DataFrame,
    path: str,
    sep="\t",
    if_exists: Literal["append", "replace", "skip"] = "replace",
    **kwargs,
) -> None:
    """
    Task to create csv file based on pandas DataFrame.
    Args:
    df (pd.DataFrame): Input pandas DataFrame.
    path (str): Path to output csv file.
    sep (str, optional): The separator to use in the CSV. Defaults to "\t".
    if_exists (Literal["append", "replace", "skip"], optional): What to do if the table exists. Defaults to "replace".
    """

    if if_exists == "append" and os.path.isfile(path):
        csv_df = pd.read_csv(path, sep=sep)
        out_df = pd.concat([csv_df, df])
    elif if_exists == "replace":
        out_df = df
    elif if_exists == "skip" and os.path.isfile(path):
        logger.info("Skipped.")
        return
    else:
        out_df = df

    # create directories if they don't exist
    try:
        if not os.path.isfile(path):
            directory = os.path.dirname(path)
            os.makedirs(directory, exist_ok=True)
    except:
        pass

    out_df.to_csv(path, index=False, sep=sep)


@task(timeout=3600)
def df_to_parquet(
    df: pd.DataFrame,
    path: str,
    if_exists: Literal["append", "replace", "skip"] = "replace",
    **kwargs,
) -> None:
    """
    Task to create parquet file based on pandas DataFrame.
    Args:
    df (pd.DataFrame): Input pandas DataFrame.
    path (str): Path to output parquet file.
    if_exists (Literal["append", "replace", "skip"], optional): What to do if the table exists. Defaults to "replace".
    """
    if if_exists == "append" and os.path.isfile(path):
        parquet_df = pd.read_parquet(path)
        out_df = pd.concat([parquet_df, df])
    elif if_exists == "replace":
        out_df = df
    elif if_exists == "skip":
        logger.info("Skipped.")
        return
    else:
        out_df = df

    # create directories if they don't exist
    try:
        if not os.path.isfile(path):
            directory = os.path.dirname(path)
            os.makedirs(directory, exist_ok=True)
    except:
        pass

    out_df.to_parquet(path, index=False, **kwargs)


@task(timeout=3600)
def union_dfs_task(dfs: List[pd.DataFrame]):
    """
    Create one DataFrame from a list of pandas DataFrames.
    Args:
        dfs (List[pd.DataFrame]): List of pandas Dataframes to concat. In case of different size of DataFrames NaN values can appear.
    """
    return pd.concat(dfs, ignore_index=True)


@task(timeout=3600)
def write_to_json(dict_, path):
    """
    Creates json file from a dictionary. Log record informs about the writing file proccess.
    Args:
        dict_ (dict): Dictionary.
        path (str): Path to local json file.
    """
    logger = prefect.context.get("logger")

    if os.path.isfile(path):
        logger.warning(f"File {path} already exists. Overwriting...")
    else:
        logger.debug(f"Writing to {path}...")

    # create parent directories if they don't exist
    Path(path).parent.mkdir(parents=True, exist_ok=True)
    with open(path, mode="w") as f:
        json.dump(dict_, f)

    logger.debug(f"Successfully wrote to {path}.")


@task(timeout=3600)
def cleanup_validation_clutter(expectations_path):
    ge_project_path = Path(expectations_path).parent
    shutil.rmtree(ge_project_path)


@task(timeout=3600)
def df_converts_bytes_to_int(df: pd.DataFrame) -> pd.DataFrame:
    logger = prefect.context.get("logger")
    logger.info("Converting bytes in dataframe columns to list of integers")
    return df.applymap(lambda x: list(map(int, x)) if isinstance(x, bytes) else x)


@task(max_retries=3, retry_delay=timedelta(seconds=10), timeout=3600)
def df_to_dataset(
    df: pd.DataFrame, partitioning_flavor="hive", format="parquet", **kwargs
) -> None:
    """
    Use `pyarrow.dataset.write_to_dataset()` to write from a pandas DataFrame to a dataset.
    This enables several data lake-specific optimizations such as parallel writes, partitioning,
    and file size (via `max_rows_per_file` parameter).

    Args:
        df (pd.DataFrame): The pandas DataFrame to write.
        partitioning_flavor (str, optional): The partitioning flavor to use. Defaults to "hive".
        format (str, optional): The dataset format. Defaults to 'parquet'.
        kwargs: Keyword arguments to be passed to `write_to_dataset()`. See
        https://arrow.apache.org/docs/python/generated/pyarrow.dataset.write_dataset.html.

    Examples:
        table = pa.Table.from_pandas(df_contact)
        base_dir = "/home/viadot/contact"
        partition_cols = ["updated_at_year", "updated_at_month", "updated_at_day"]

        df_to_dataset(
            data=table,
            base_dir=base_dir,
            partitioning=partition_cols,
            existing_data_behavior='overwrite_or_ignore',
            max_rows_per_file=100_000
        )
    """
    table = pa.Table.from_pandas(df)
    ds.write_dataset(
        data=table, partitioning_flavor=partitioning_flavor, format=format, **kwargs
    )


@curry
def custom_mail_state_handler(
    tracked_obj: Union["Flow", "Task"],
    old_state: prefect.engine.state.State,
    new_state: prefect.engine.state.State,
    only_states: list = [Failed],
    local_api_key: str = None,
    credentials_secret: str = None,
    vault_name: str = None,
    from_email: str = None,
    to_emails: str = None,
) -> prefect.engine.state.State:
    """
    Custom state handler configured to work with sendgrid.
    Works as a standalone state handler, or can be called from within a custom state handler.
    Args:
        tracked_obj (Task or Flow): Task or Flow object the handler is registered with.
        old_state (State): previous state of tracked object.
        new_state (State): new state of tracked object.
        only_states ([State], optional): similar to `ignore_states`, but instead _only_
            notifies you if the Task / Flow is in a state from the provided list of `State`
            classes.
        local_api_key (str, optional): Api key from local config.
        credentials_secret (str, optional): The name of the Azure Key Vault secret containing a dictionary with API KEY.
        vault_name (str, optional): Name of key vault.
        from_email (str): Sender mailbox address.
        to_emails (str): Receiver mailbox address.
    Returns: State: the `new_state` object that was provided

    """

    if credentials_secret is None:
        try:
            credentials_secret = PrefectSecret("SENDGRID_DEFAULT_SECRET").run()
        except ValueError:
            pass

    if credentials_secret is not None:
        credentials_str = AzureKeyVaultSecret(
            credentials_secret, vault_name=vault_name
        ).run()
        api_key = json.loads(credentials_str).get("API_KEY")
    elif local_api_key is not None:
        api_key = local_config.get(local_api_key).get("API_KEY")
    else:
        raise Exception("Please provide API KEY")

    curr_dt = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    only_states = only_states or []
    if only_states and not any(
        [isinstance(new_state, included) for included in only_states]
    ):
        return new_state
    url = prefect.client.Client().get_cloud_url(
        "flow-run", prefect.context["flow_run_id"], as_user=False
    )
    message = Mail(
        from_email=from_email,
        to_emails=to_emails,
        subject=f"The flow {tracked_obj.name} - Status {new_state}",
        html_content=f"<strong>The flow {cast(str,tracked_obj.name)} FAILED at {curr_dt}. \
    <p>More details here: {url}</p></strong>",
    )
    try:
        send_grid = SendGridAPIClient(api_key)
        response = send_grid.send(message)
    except Exception as e:
        raise e

    return new_state


@task(timeout=3600)
def df_clean_column(
    df: pd.DataFrame, columns_to_clean: List[str] = None
) -> pd.DataFrame:
    """
    Function that removes special characters (such as escape symbols)
    from a pandas DataFrame.

    Args:
    df (pd.DataFrame): The DataFrame to clean.
    columns_to_clean (List[str]): A list of columns to clean. Defaults is None.

    Returns:
    pd.DataFrame: The cleaned DataFrame.
    """

    df = df.copy()
    logger.info(f"Removing special characters from dataframe columns...")

    if columns_to_clean is None:
        df.replace(
            to_replace=[r"\\t|\\n|\\r", "\t|\n|\r"],
            value=["", ""],
            regex=True,
            inplace=True,
        )
    else:
        for col in columns_to_clean:
            df[col].replace(
                to_replace=[r"\\t|\\n|\\r", "\t|\n|\r"],
                value=["", ""],
                regex=True,
                inplace=True,
            )
    return df


@task(timeout=3600)
def concat_dfs(dfs: List[pd.DataFrame]):
    """
    Task to combine list of data frames into one.

    Args:
        dfs (List[pd.DataFrame]): List of dataframes to concat.
    Returns:
        pd.DataFrame(): Pandas dataframe containing all columns from dataframes from list.
    """
    return pd.concat(dfs, axis=1)


@task(timeout=3600)
def cast_df_to_str(df: pd.DataFrame) -> pd.DataFrame:
    """
    Task for casting an entire DataFrame to a string data type. Task is needed
    when data is being uploaded from Parquet file to DuckDB because empty columns
    can be casted to INT instead of default VARCHAR.

    Args:
        df (pd.DataFrame): Input DataFrame.

    Returns:
        df_mapped (pd.DataFrame): Pandas DataFrame casted to string.
    """
    df_mapped = df.astype("string")
    return df_mapped


@task(timeout=3600)
def set_new_kv(kv_name: str, df: pd.DataFrame, filter_column: str):
    """
    Task for updating/setting key value on Prefect based on the newest
    values in pandas DataFrame.

    Args:
        kv_name (str): Name of key value to change.
        df (pd.DataFrame): DataFrame based on which value will be updated.
        filter_column (str): Field from which obtain new value.
    """
    if df.empty:
        logger.warning("Input DataFrame is empty. Cannot set a new key value.")
    else:
        new_value = str(df[filter_column].max()).strip()
        set_key_value(key=kv_name, value=new_value)


class Git(Git):
    @property
    def git_clone_url(self):
        """
        Build the git url to clone.
        """
        if self.use_ssh:
            return f"git@{self.repo_host}:{self.repo}"
        return f"https://{self.git_token_secret}@{self.repo_host}/{self.repo}"


@task(timeout=3600)
def credentials_loader(credentials_secret: str, vault_name: str = None) -> dict:
    """
    Function that gets credentials from azure Key Vault or PrefectSecret or from local config.

    Args:
        credentials_secret (str): The name of the Azure Key Vault secret containing a dictionary
        with credentials.
        vault_name (str, optional): The name of the vault from which to obtain the secret. Defaults to None.

    Returns:
        credentials (dict): Credentials as dictionary.
    """

    if credentials_secret:
        try:
            credentials = local_config[credentials_secret]
            logger.info("Successfully loaded credentials from local config.")
        except (ValueError, KeyError):
            try:
                azure_secret_task = AzureKeyVaultSecret()
                credentials_str = azure_secret_task.run(
                    secret=credentials_secret, vault_name=vault_name
                )
                credentials = json.loads(credentials_str)
                logger.info("Successfully loaded credentials from Azure Key Vault.")
            except Exception:
                try:
                    credentials = PrefectSecret(credentials_secret).run()
                    logger.info("Successfully loaded credentials from PrefectSecret.")
                except Exception:
                    raise CredentialError(
                        "Provided credentials secret not found in resources."
                    )
    else:
        raise CredentialError("Credentials secret not provided.")

    return credentials


@task(timeout=3600)
def adls_bulk_upload(
    file_names: List[str],
    file_name_relative_path: str = "",
    adls_file_path: str = None,
    adls_sp_credentials_secret: str = None,
    adls_overwrite: bool = True,
    timeout: int = 3600,
) -> None:
    """Function that upload files to defined path in ADLS.
    Args:
        file_names (List[str]): List of file names to generate paths.
        file_name_relative_path (str, optional): Path where to save the file locally. Defaults to ''.
        adls_file_path (str, optional): Azure Data Lake path. Defaults to None.
        adls_sp_credentials_secret (str, optional): The name of the Azure Key Vault secret containing a dictionary with
            ACCOUNT_NAME and Service Principal credentials (TENANT_ID, CLIENT_ID, CLIENT_SECRET). Defaults to None.
        adls_overwrite (bool, optional): Whether to overwrite files in the data lake. Defaults to True.
        timeout (int, optional): The amount of time (in seconds) to wait while running this task before
            a timeout occurs. Defaults to 3600.
    """

    file_to_adls_task = AzureDataLakeUpload(timeout=timeout)

    for file in file_names:
        file_to_adls_task.run(
            from_path=os.path.join(file_name_relative_path, file),
            to_path=os.path.join(adls_file_path, file),
            sp_credentials_secret=adls_sp_credentials_secret,
            overwrite=adls_overwrite,
        )


@task(timeout=3600)
def anonymize_df(
    df: pd.DataFrame,
    columns: List[str],
    method: Literal["mask", "hash"] = "mask",
    value: str = "***",
    date_column: str = None,
    days: int = None,
) -> pd.DataFrame:
    """
    Function that anonymize data in the dataframe in selected columns.
    It is possible to specify the condtition, for which data older than specified number of days will be anonymized.

    Args:
        df (pd.DataFrame): Dataframe with data to anonymize.
        columns (List[str]): List of columns to anonymize.
        method (Literal["mask", "hash"], optional): Method of anonymizing data. "mask" -> replace the data with "value" arg.
            "hash" -> replace the data with the hash value of an object (using `hash()` method). Defaults to "mask".
        value (str, optional): Value to replace the data. Defaults to "***".
        date_column (str, optional): Name of the date column used to identify rows that are older than a specified number of days. Defaults to None.
        days (int, optional): The number of days beyond which we want to anonymize the data, e.g. older that 2 years can be: 2*365. Defaults to None.

    Examples:
        1. Implement "mask" method with "***" for all data in columns: ["email", "last_name", "phone"]:
            >>> anonymize_df(df=df, columns=["email", "last_name", "phone"])
        2. Implement "hash" method with in columns: ["email", "last_name", "phone"]:
            >>> anonymize_df(df=df, columns=["email", "last_name", "phone"], method = "hash")
        3. Implement "mask" method with "***" for data in columns: ["email", "last_name", "phone"], that is older than two years in "submission_date" column:
            >>> anonymize_df(df=df, columns=["email", "last_name", "phone"], date_column="submission_date", days=2*365)

    Raises:
        ValueError: If method or columns not found.

    Returns:
        pd.DataFrame: Operational dataframe with anonymized data.
    """
    if all(col in df.columns for col in columns) == False:
        raise ValueError(
            f"At least one of the following columns is not found in dataframe: {columns} or argument is not list. Provide list with proper column names."
        )

    if days and date_column:
        days_ago = datetime.now().date() - timedelta(days=days)
        df["temp_date_col"] = pd.to_datetime(df[date_column]).dt.date

        to_hash = df["temp_date_col"] < days_ago
        if any(to_hash) == False:
            logger.warning(f"No data that is older than {days} days.")
        else:
            logger.info(
                f"Data older than {days} days in {columns} columns will be anonymized."
            )
    else:
        to_hash = len(df.index) * [True]
        logger.info(
            f"The 'days' and 'date_column' arguments were not specified. All data in {columns} columns will be anonymized."
        )

    if method == "mask":
        df.loc[to_hash, columns] = value
    elif method == "hash":
        df.loc[to_hash, columns] = df.loc[to_hash, columns].apply(
            lambda x: x.apply(hash)
        )
    else:
        raise ValueError(
            f"Method not found. Use one of the available methods: 'mask', 'hash'."
        )

    df.drop(columns=["temp_date_col"], inplace=True, errors="ignore")
    return df



import json

import pytest
from prefect.tasks.secrets import PrefectSecret

from viadot.flows.sql_server_to_duckdb import SQLServerToDuckDB
from viadot.tasks import AzureSQLDBQuery, DuckDBQuery, DuckDBToDF, SQLServerCreateTable
from viadot.tasks.azure_key_vault import AzureKeyVaultSecret

SCHEMA = "sandbox"
TABLE = "test"


@pytest.fixture(scope="session")
def create_sql_server_table():
    credentials_secret = PrefectSecret(
        "AZURE_DEFAULT_SQLDB_SERVICE_PRINCIPAL_SECRET"
    ).run()
    vault_name = PrefectSecret("AZURE_DEFAULT_KEYVAULT").run()
    azure_secret_task = AzureKeyVaultSecret()
    credentials_str = azure_secret_task.run(
        secret=credentials_secret, vault_name=vault_name
    )
    dtypes = {
        "date": "DATE",
        "name": "VARCHAR(255)",
        "id": "VARCHAR(255)",
        "weather": "FLOAT(24)",
        "rain": "FLOAT(24)",
        "temp": "FLOAT(24)",
        "summary": "VARCHAR(255)",
    }
    create_table_task = SQLServerCreateTable()
    yield create_table_task.run(
        schema=SCHEMA,
        table=TABLE,
        dtypes=dtypes,
        if_exists="replace",
        credentials=json.loads(credentials_str),
    )
    drop_sqlserver = AzureSQLDBQuery()
    drop_sqlserver.run(
        query=f"DROP TABLE {SCHEMA}.{TABLE}", credentials_secret=credentials_secret
    )


def test_sql_server_to_duckdb(create_sql_server_table):
    create_sql_server_table
    duckdb_creds = {"database": "/home/viadot/database/test.duckdb"}
    flow = SQLServerToDuckDB(
        name="test",
        sql_query=f"SELECT * FROM {SCHEMA}.{TABLE}",
        local_file_path="test.parquet",
        sqlserver_config_key="AZURE_SQL",
        if_exists="replace",
        duckdb_table=TABLE,
        duckdb_schema=SCHEMA,
        duckdb_credentials=duckdb_creds,
    )
    result = flow.run()
    assert result.is_successful()

    df_task = DuckDBToDF(credentials=duckdb_creds)
    df = df_task.run(table=TABLE, schema=SCHEMA)

    assert df.columns.to_list() == [
        "date",
        "name",
        "id",
        "weather",
        "rain",
        "temp",
        "summary",
        "_viadot_downloaded_at_utc",
    ]
    drop_duckdb = DuckDBQuery()
    drop_duckdb.run(query=f"DROP TABLE {SCHEMA}.{TABLE}", credentials=duckdb_creds)



import json
from typing import List

import pandas as pd
import pytest
from prefect.tasks.secrets import PrefectSecret

from viadot.exceptions import APIError
from viadot.sources import Mediatool
from viadot.task_utils import credentials_loader

CREDENTIALS = credentials_loader.run(credentials_secret="MEDIATOOL-TESTS")
MTOOL = Mediatool(credentials=CREDENTIALS)


def test_get_campaigns_df():
    camps = MTOOL.get_campaigns(CREDENTIALS["ORG"])
    assert isinstance(camps, pd.DataFrame)


def test_get_campaigns_dict():
    camps = MTOOL.get_campaigns(CREDENTIALS["ORG"], return_dataframe=False)
    assert isinstance(camps, list)
    assert isinstance(camps[0], dict)


def test_get_organizations():
    orgs = MTOOL.get_organizations(user_id=CREDENTIALS["USER_ID"])
    assert isinstance(orgs, pd.DataFrame)


def test_get_media_entries():
    media_entries = MTOOL.get_media_entries(
        organization_id=CREDENTIALS["ORG"], columns=["_id"]
    )
    assert isinstance(media_entries, pd.DataFrame)


def test_get_media_types_correct_id():
    media_types = MTOOL.get_media_types(media_type_ids=[CREDENTIALS["MEDIA_TYPE_ID"]])
    assert isinstance(media_types, pd.DataFrame)


def test_get_media_types_wrong_id():
    with pytest.raises(
        APIError, match=r"Perhaps your account credentials need to be refreshed"
    ):
        _ = MTOOL.get_media_types(["040404"])


def test_get_vehicles(caplog):
    _ = MTOOL.get_vehicles(vehicle_ids=["100000", "200000"])
    assert "Vehicle were not found for: ['100000', '200000']" in caplog.text


def test_rename_columns_correct():
    data = {"id": [1, 2], "amount": [3, 4]}
    df = pd.DataFrame(data=data)
    df_updated = MTOOL.rename_columns(df=df, column_suffix="source_test")

    assert isinstance(df_updated, pd.DataFrame)
    assert "id_source_test" in df_updated.columns


def test_rename_columns_not_df():
    with pytest.raises(TypeError, match=r"is not a data frame."):
        _ = MTOOL.rename_columns(df="DF", column_suffix="source_test")


def test_rename_columns_empty_source_name():
    data = {"id": [1, 2], "amount": [3, 4]}
    df = pd.DataFrame(data=data)
    df_updated = MTOOL.rename_columns(df=df)

    assert isinstance(df_updated, pd.DataFrame)
    assert "id_rename" in df_updated.columns



from typing import Any, Dict, List, Literal

import pandas as pd
from prefect import Task
from prefect.utilities import logging
from prefect.utilities.tasks import defaults_from_attrs

from viadot.task_utils import *

from ..sources import Mediatool

logger = logging.get_logger()


class MediatoolToDF(Task):
    """
    The MediatoolToDF task fetches data for media entries, combines them with other data from Mediatool endpoints,
    and returns data with names instead of IDs. User can provide list of organization IDs.
    """

    def __init__(
        self,
        organization_ids: List[str] = None,
        media_entries_columns: List[str] = None,
        mediatool_credentials: dict = None,
        mediatool_credentials_key: str = None,
        *args: List[Any],
        **kwargs: Dict[str, Any],
    ):
        """
        Task MediatoolToDF for joining all of the data for media_entries and return as dataframe.

        Args:
            organization_ids (List[str]): List of organization IDs.
            media_entries_columns (List[str], optional): List of media entries fields to download. Defaults to None.
            mediatool_credentials (dict, optional): Dictionary containing Mediatool credentials. Defaults to None.
            mediatool_credentials_key (str, optional): Key for Mediatool credentials. Defaults to None.
        """

        self.media_entries_columns = media_entries_columns
        self.organization_ids = organization_ids

        if mediatool_credentials is None:
            self.mediatool_credentials = credentials_loader.run(
                credentials_secret=mediatool_credentials_key,
            )

        else:
            self.mediatool_credentials = mediatool_credentials

        super().__init__(
            name="mediatool_to_df",
            *args,
            **kwargs,
        )

    def join_dfs(
        self,
        df_left: pd.DataFrame,
        df_right: pd.DataFrame,
        left_on: str,
        right_on: str,
        columns_from_right_df: List[str] = None,
        how: Literal["left", "right", "outer", "inner", "cross"] = "left",
    ) -> pd.DataFrame:
        """
        Combine the dataframes according to the chosen method.

        Args:
            df_left (pd.DataFrame): Left dataframe.
            df_right (pd.DataFrame): Right dataframe.
            left_on (str): Column or index level names to join on in the left DataFrame.
            right_on (str): Column or index level names to join on in the right DataFrame.
            columns_from_right_df (List[str], optional): List of column to get from right dataframe.
                Defaults to None.
            how (Literal["left", "right", "outer", "inner", "cross"], optional): Type of merge to be performed.
                Defaults to "left".

        Returns:
            pd.DataFrame: Final dataframe after merging.
        """
        if columns_from_right_df is None:
            columns_from_right_df = df_right.columns

        df_merged = df_left.merge(
            df_right[columns_from_right_df],
            left_on=left_on,
            right_on=right_on,
            how=how,
        )
        return df_merged

    def __call__(self, *args, **kwargs):
        """Download Mediatool data to DF."""
        return super().__call__(*args, **kwargs)

    @defaults_from_attrs("organization_ids", "media_entries_columns")
    def run(
        self, organization_ids: List[str] = None, media_entries_columns=None
    ) -> pd.DataFrame:
        """
        Run method. Data for different endpoints are retrieved and combined.
        Finally, a final data frame is created containing data for all organizations from the list.

        Args:
            organization_ids (List[str], optional): List of organization IDs. Defaults to None.
            media_entries_columns (List[str], optional): Columns to get from media entries. Defaults to None.

        Raises:
            ValueError: Value is raised if organization_id is not recognized.

        Returns:
            pd.DataFrame: Data frame containing all of the information for organization or list of organizations.
        """
        mediatool = Mediatool(credentials=self.mediatool_credentials)
        df_orgs = mediatool.get_organizations(self.mediatool_credentials["USER_ID"])

        list_of_dfs = []
        for organization_id in organization_ids:
            if organization_id in df_orgs["_id_organizations"].unique():
                logger.info(f"Downloading data for: {organization_id} ...")

                # extract data
                df_media_entries = mediatool.get_media_entries(
                    organization_id=organization_id,
                    columns=media_entries_columns,
                )
                df_camp = mediatool.get_campaigns(organization_id=organization_id)

                unique_vehicle_ids = df_media_entries["vehicleId"].unique()
                df_veh = mediatool.get_vehicles(vehicle_ids=unique_vehicle_ids)

                unique_media_type_ids = df_media_entries["mediaTypeId"].unique()
                df_m_types = mediatool.get_media_types(unique_media_type_ids)

                # join DFs
                df_merged_entries_orgs = self.join_dfs(
                    df_left=df_media_entries,
                    df_right=df_orgs,
                    left_on="organizationId",
                    right_on="_id_organizations",
                    columns_from_right_df=[
                        "_id_organizations",
                        "name_organizations",
                        "abbreviation_organizations",
                    ],
                    how="left",
                )

                df_merged_campaigns = self.join_dfs(
                    df_left=df_merged_entries_orgs,
                    df_right=df_camp,
                    left_on="campaignId",
                    right_on="_id_campaigns",
                    columns_from_right_df=[
                        "_id_campaigns",
                        "name_campaigns",
                        "conventionalName_campaigns",
                    ],
                    how="left",
                )

                df_merged_vehicles = self.join_dfs(
                    df_left=df_merged_campaigns,
                    df_right=df_veh,
                    left_on="vehicleId",
                    right_on="_id_vehicles",
                    columns_from_right_df=["_id_vehicles", "name_vehicles"],
                    how="left",
                )

                df_merged_media_types = self.join_dfs(
                    df_left=df_merged_vehicles,
                    df_right=df_m_types,
                    left_on="mediaTypeId",
                    right_on="_id_media_types",
                    columns_from_right_df=["_id_media_types", "name_media_types"],
                    how="left",
                )

                list_of_dfs.append(df_merged_media_types)

            else:
                raise ValueError(
                    f"Organization - {organization_id} not found in organizations list."
                )

            df_final = pd.concat(list_of_dfs)

        return df_final



import json
from datetime import timedelta
from typing import Dict, Generator, List

import pandas as pd
from prefect import Task
from prefect.tasks.secrets import PrefectSecret
from prefect.utilities.tasks import defaults_from_attrs

from viadot.config import local_config

from ..sources import CloudForCustomers
from .azure_key_vault import AzureKeyVaultSecret


class C4CReportToDF(Task):
    def __init__(
        self,
        *args,
        report_url: str = None,
        skip: int = 0,
        top: int = 1000,
        env: str = "QA",
        max_retries: int = 3,
        retry_delay: timedelta = timedelta(seconds=10),
        timeout: int = 3600,
        **kwargs,
    ):
        self.report_url = report_url
        self.env = env
        self.skip = skip
        self.top = top

        super().__init__(
            name="c4c_report_to_df",
            max_retries=max_retries,
            retry_delay=retry_delay,
            timeout=timeout,
            *args,
            **kwargs,
        )

    def __call__(self, *args, **kwargs):
        """Download report to DF"""
        return super().__call__(*args, **kwargs)

    @defaults_from_attrs(
        "report_url",
        "env",
        "skip",
        "top",
    )
    def run(
        self,
        report_url: str = None,
        env: str = "QA",
        skip: int = 0,
        top: int = 1000,
        credentials_secret: str = None,
        vault_name: str = None,
        max_retries: int = 3,
        retry_delay: timedelta = timedelta(seconds=10),
    ):
        """
        Task for downloading data from the Cloud for Customers to a pandas DataFrame using report URL.

        Args:
            report_url (str, optional): The URL to the report. Defaults to None.
            env (str, optional): The environment to use. Defaults to 'QA'.
            skip (int, optional): Initial index value of reading row. Defaults to 0.
            top (int, optional): The value of top reading row. Defaults to 1000.
            credentials_secret (str, optional): The name of the Azure Key Vault secret containing a dictionary
            with C4C credentials (username & password). Defaults to None.
            vault_name (str, optional): The name of the vault from which to obtain the secret. Defaults to None.

        Returns:
            pd.DataFrame: The query result as a pandas DataFrame.
        """

        if not credentials_secret:
            try:
                credentials_secret = PrefectSecret("C4C_KV").run()
            except ValueError:
                pass

        if credentials_secret:
            credentials_str = AzureKeyVaultSecret(
                credentials_secret, vault_name=vault_name
            ).run()
            credentials = json.loads(credentials_str)
        else:
            credentials = local_config.get("CLOUD_FOR_CUSTOMERS")[env]

        final_df = pd.DataFrame()
        next_batch = True
        while next_batch:
            new_url = f"{report_url}&$top={top}&$skip={skip}"
            chunk_from_url = CloudForCustomers(
                report_url=new_url, env=env, credentials=credentials
            )
            df = chunk_from_url.to_df()
            final_df = final_df.append(df)
            if not final_df.empty:
                df_count = df.count()[1]
                if df_count != top:
                    next_batch = False
                skip += top
            else:
                break
        return final_df


class C4CToDF(Task):
    def __init__(
        self,
        *args,
        url: str = None,
        endpoint: str = None,
        fields: List[str] = None,
        params: Dict[str, str] = None,
        chunksize: int = 20000,
        env: str = "QA",
        if_empty: str = "warn",
        max_retries: int = 3,
        retry_delay: timedelta = timedelta(seconds=10),
        timeout: int = 3600,
        **kwargs,
    ):
        self.url = url
        self.endpoint = endpoint
        self.fields = fields
        self.params = params
        self.chunksize = chunksize
        self.env = env
        self.if_empty = if_empty

        super().__init__(
            name="c4c_to_df",
            max_retries=max_retries,
            retry_delay=retry_delay,
            timeout=timeout,
            *args,
            **kwargs,
        )

    @defaults_from_attrs(
        "url", "endpoint", "fields", "params", "chunksize", "env", "if_empty"
    )
    def run(
        self,
        url: str = None,
        env: str = "QA",
        endpoint: str = None,
        fields: List[str] = None,
        params: Dict[str, str] = None,
        chunksize: int = None,
        if_empty: str = "warn",
        credentials_secret: str = None,
        vault_name: str = None,
    ):
        """
        Task for downloading data from the Cloud for Customers to a pandas DataFrame using URL (with query parameters).
        Data is obtained from table by passing table name in the url or endpoint. It is recommended to add filters
        parameters in this case.

        Example:
            url = "https://mysource.com/sap/c4c/odata/v1/c4codataapi"
            endpoint = "ServiceRequestCollection"
            params = {"$filter": "CreationDateTime ge 2021-12-21T00:00:00Z"}

        Args:
            url (str, optional): The url to the API used in case of prepared report. Defaults to None.
            env (str, optional): The environment to use to obtain credentials. Defaults to 'QA'.
            endpoint (str, optional): The endpoint of the API. Defaults to None.
            fields (List[str], optional): The C4C Table fields. Defaults to None.
            params (Dict[str, str]): Query parameters. Defaults to None.
            chunksize (int, optional): How many rows to retrieve from C4C at a time. Uses a server-side cursor. Defaults to None.
            if_empty (str, optional): What to do if query returns no data. Defaults to "warn".
            credentials_secret (str, optional): The name of the Azure Key Vault secret containing a dictionary
            with C4C credentials. Defaults to None.
            vault_name (str, optional): The name of the vault from which to obtain the secret. Defaults to None.

        Returns:
            pd.DataFrame: The query result as a pandas DataFrame.
        """
        if not credentials_secret:
            try:
                credentials_secret = PrefectSecret("C4C_KV").run()
            except ValueError:
                pass

        if credentials_secret:
            credentials_str = AzureKeyVaultSecret(
                credentials_secret, vault_name=vault_name
            ).run()
            credentials = json.loads(credentials_str)
        else:
            credentials = local_config.get("CLOUD_FOR_CUSTOMERS")[env]

        self.logger.info(f"Downloading data from {url+endpoint}...")

        # If we get any of these in params, we don't perform any chunking
        if any(["$skip" in params, "$top" in params]):
            return CloudForCustomers(
                url=url,
                endpoint=endpoint,
                params=params,
                env=env,
                credentials=credentials,
            ).to_df(if_empty=if_empty, fields=fields)

        def _generate_chunks() -> Generator[pd.DataFrame, None, None]:
            """
            Util function returning chunks.

            Returns: Generator[pd.DataFrame, None, None]
            """
            offset = 0
            total_record_count = 0
            while True:
                boundaries = {"$skip": offset, "$top": chunksize}
                params.update(boundaries)

                chunk = CloudForCustomers(
                    url=url,
                    endpoint=endpoint,
                    params=params,
                    env=env,
                    credentials=credentials,
                ).to_df(if_empty=if_empty, fields=fields)

                chunk_record_count = chunk.shape[0]
                total_record_count += chunk_record_count
                self.logger.info(
                    f"Successfully downloaded {total_record_count} records."
                )

                yield chunk

                if chunk.shape[0] < chunksize:
                    break

                offset += chunksize

        self.logger.info(f"Data from {url+endpoint} has been downloaded successfully.")

        chunks = _generate_chunks()
        df = pd.concat(chunks)

        return df



import json
from asyncio.log import logger
from datetime import timedelta
from typing import Any, Dict, List, Literal

import pandas as pd
from prefect import Task
from prefect.tasks.secrets import PrefectSecret
from prefect.utilities.tasks import defaults_from_attrs

from ..exceptions import ValidationError
from ..sources import AzureSQL
from ..utils import (
    build_merge_query,
    gen_bulk_insert_query_from_df,
    get_sql_server_table_dtypes,
)
from .azure_key_vault import AzureKeyVaultSecret


def get_credentials(credentials_secret: str, vault_name: str = None):
    """
    Get Azure credentials.
    If the credential secret is not provided it will be taken from Prefect Secrets. If Prefect Secrets does not
        contain the credential, it will be taken from the local credential file.

    Args:
        credentials_secret (str): The name of the Azure Key Vault secret containing a dictionary
        with SQL db credentials (server, db_name, user and password).
        vault_name (str, optional): The name of the vault from which to obtain the secret. Defaults to None.

    Returns: Credentials

    """
    if not credentials_secret:
        # attempt to read a default for the service principal secret name
        try:
            credentials_secret = PrefectSecret(
                "AZURE_DEFAULT_SQLDB_SERVICE_PRINCIPAL_SECRET"
            ).run()
        except ValueError:
            pass

    if credentials_secret:
        azure_secret_task = AzureKeyVaultSecret()
        credentials_str = azure_secret_task.run(
            secret=credentials_secret, vault_name=vault_name
        )
        credentials = json.loads(credentials_str)

        return credentials


class CreateTableFromBlob(Task):
    def __init__(self, sep="\t", timeout: int = 3600, *args, **kwargs):
        self.sep = sep
        super().__init__(name="blob_to_azure_sql", timeout=timeout, *args, **kwargs)

    @defaults_from_attrs("sep")
    def run(
        self,
        blob_path: str,
        schema: str,
        table: str,
        dtypes: Dict[str, Any],
        sep: str = None,
        if_exists: Literal = ["fail", "replace", "append", "delete"],
    ):
        """
        Create a table from an Azure Blob object.
        Currently, only CSV files are supported.

        Args:
        from_path (str): Path to the file to be inserted.
        schema (str): Destination schema.
        table (str): Destination table.
        dtypes (Dict[str, Any]): Data types to force.
        sep (str): The separator to use to read the CSV file.
        if_exists (Literal, optional): What to do if the table already exists.
        """

        fqn = f"{schema}.{table}" if schema else table
        azure_sql = AzureSQL(config_key="AZURE_SQL")

        if if_exists == "replace":
            azure_sql.create_table(
                schema=schema, table=table, dtypes=dtypes, if_exists=if_exists
            )

            self.logger.info(f"Successfully created table {fqn}.")

        azure_sql.bulk_insert(
            schema=schema,
            table=table,
            source_path=blob_path,
            sep=sep,
            if_exists=if_exists,
        )
        self.logger.info(f"Successfully inserted data into {fqn}.")


class AzureSQLBulkInsert(Task):
    def __init__(
        self,
        from_path: str = None,
        schema: str = None,
        table: str = None,
        dtypes: Dict[str, Any] = None,
        sep="\t",
        if_exists: Literal["fail", "replace", "append", "delete"] = "fail",
        credentials_secret: str = None,
        timeout: int = 3600,
        *args,
        **kwargs,
    ):
        self.from_path = from_path
        self.schema = schema
        self.table = table
        self.dtypes = dtypes
        self.sep = sep
        self.if_exists = if_exists
        self.credentials_secret = credentials_secret
        super().__init__(name="azure_sql_bulk_insert", timeout=timeout, *args, **kwargs)

    @defaults_from_attrs("sep", "if_exists", "credentials_secret")
    def run(
        self,
        from_path: str = None,
        schema: str = None,
        table: str = None,
        dtypes: Dict[str, Any] = None,
        sep: str = None,
        if_exists: Literal["fail", "replace", "append", "delete"] = None,
        credentials_secret: str = None,
        vault_name: str = None,
    ):
        """
        Bulk insert data from Azure Data Lake into an Azure SQL Database table.
        This task also creates the table if it doesn't exist.
        Currently, only CSV files are supported.

        Args:
        from_path (str): Path to the file to be inserted.
        schema (str): Destination schema.
        table (str): Destination table.
        dtypes (Dict[str, Any]): Data types to force.
        sep (str): The separator to use to read the CSV file.
        if_exists (Literal, optional): What to do if the table already exists.
        credentials_secret (str, optional): The name of the Azure Key Vault secret containing a dictionary
        with SQL db credentials (server, db_name, user, and password).
        vault_name (str, optional): The name of the vault from which to obtain the secret. Defaults to None.
        """

        fqn = f"{schema}.{table}" if schema else table
        credentials = get_credentials(credentials_secret, vault_name=vault_name)
        azure_sql = AzureSQL(credentials=credentials)

        if if_exists == "replace":
            azure_sql.create_table(
                schema=schema, table=table, dtypes=dtypes, if_exists=if_exists
            )

            self.logger.info(f"Successfully created table {fqn}.")

        azure_sql.bulk_insert(
            schema=schema,
            table=table,
            source_path=from_path,
            sep=sep,
            if_exists=if_exists,
        )
        self.logger.info(f"Successfully inserted data into {fqn}.")


class AzureSQLCreateTable(Task):
    def __init__(
        self,
        schema: str = None,
        table: str = None,
        dtypes: Dict[str, Any] = None,
        if_exists: Literal["fail", "replace", "skip", "delete"] = "fail",
        credentials_secret: str = None,
        vault_name: str = None,
        timeout: int = 3600,
        max_retries: int = 3,
        retry_delay: timedelta = timedelta(seconds=10),
        *args,
        **kwargs,
    ):
        self.schema = schema
        self.table = table
        self.dtypes = dtypes
        self.if_exists = if_exists
        self.credentials_secret = credentials_secret
        self.vault_name = vault_name
        super().__init__(
            name="azure_sql_create_table",
            max_retries=max_retries,
            retry_delay=retry_delay,
            timeout=timeout,
            *args,
            **kwargs,
        )

    @defaults_from_attrs("if_exists")
    def run(
        self,
        schema: str = None,
        table: str = None,
        dtypes: Dict[str, Any] = None,
        if_exists: Literal["fail", "replace", "skip", "delete"] = None,
        credentials_secret: str = None,
        vault_name: str = None,
        max_retries: int = None,
        retry_delay: timedelta = None,
    ):
        """
        Create a table in Azure SQL Database.

        Args:
            schema (str, optional): Destination schema.
            table (str, optional): Destination table.
            dtypes (Dict[str, Any], optional): Data types to force.
            if_exists (Literal, optional): What to do if the table already exists.
            credentials_secret (str, optional): The name of the Azure Key Vault secret containing a dictionary
            with SQL db credentials (server, db_name, user, and password).
            vault_name (str, optional): The name of the vault from which to obtain the secret. Defaults to None.
        """

        credentials = get_credentials(credentials_secret, vault_name=vault_name)
        azure_sql = AzureSQL(credentials=credentials)

        fqn = f"{schema}.{table}" if schema is not None else table
        created = azure_sql.create_table(
            schema=schema, table=table, dtypes=dtypes, if_exists=if_exists
        )
        if created:
            self.logger.info(f"Successfully created table {fqn}.")
        else:
            self.logger.info(
                f"Table {fqn} has not been created as if_exists is set to {if_exists}."
            )


class AzureSQLDBQuery(Task):
    """
    Task for running an Azure SQL Database query.

    Args:
        query (str, required): The query to execute on the database.
        credentials_secret (str, optional): The name of the Azure Key Vault secret containing a dictionary
        with SQL db credentials (server, db_name, user, and password).
        vault_name (str, optional): The name of the vault from which to obtain the secret. Defaults to None.
        timeout(int, optional): The amount of time (in seconds) to wait while running this task before
            a timeout occurs. Defaults to 3600.
    """

    def __init__(
        self,
        credentials_secret: str = None,
        vault_name: str = None,
        timeout: int = 3600,
        *args,
        **kwargs,
    ):
        self.credentials_secret = credentials_secret
        self.vault_name = vault_name

        super().__init__(name="azure_sql_db_query", timeout=timeout, *args, **kwargs)

    def run(
        self,
        query: str,
        credentials_secret: str = None,
        vault_name: str = None,
    ):
        """Run an Azure SQL Database query

        Args:
            query (str, required): The query to execute on the database.
            credentials_secret (str, optional): The name of the Azure Key Vault secret containing a dictionary
            with SQL db credentials (server, db_name, user, and password).
            vault_name (str, optional): The name of the vault from which to obtain the secret. Defaults to None.
        """

        credentials = get_credentials(credentials_secret, vault_name=vault_name)
        azure_sql = AzureSQL(credentials=credentials)

        # run the query and fetch the results if it's a select
        result = azure_sql.run(query)

        self.logger.info(f"Successfully ran the query.")
        return result


class AzureSQLToDF(Task):
    """
    Task for loading the result of an Azure SQL Database query into a pandas DataFrame.

    Args:
        query (str, required): The query to execute on the database.
        credentials_secret (str, optional): The name of the Azure Key Vault secret containing a dictionary
        with SQL db credentials (server, db_name, user, and password).
        vault_name (str, optional): The name of the vault from which to obtain the secret. Defaults to None.
        timeout(int, optional): The amount of time (in seconds) to wait while running this task before
            a timeout occurs. Defaults to 3600.
    """

    def __init__(
        self,
        credentials_secret: str = None,
        vault_name: str = None,
        timeout: int = 3600,
        *args,
        **kwargs,
    ):
        self.credentials_secret = credentials_secret
        self.vault_name = vault_name

        super().__init__(name="azure_sql_to_df", timeout=timeout, *args, **kwargs)

    def run(
        self,
        query: str,
        credentials_secret: str = None,
        vault_name: str = None,
    ):
        """Load the result of an Azure SQL Database query into a pandas DataFrame.

        Args:
            query (str, required): The query to execute on the database.
            credentials_secret (str, optional): The name of the Azure Key Vault secret containing a dictionary
            with SQL db credentials (server, db_name, user, and password).
            vault_name (str, optional): The name of the vault from which to obtain the secret. Defaults to None.
        """

        credentials = get_credentials(credentials_secret, vault_name=vault_name)
        azure_sql = AzureSQL(credentials=credentials)

        df = azure_sql.to_df(query)
        nrows = df.shape[0]
        ncols = df.shape[1]

        self.logger.info(
            f"Successfully downloaded {nrows} rows and {ncols} columns of data to a DataFrame."
        )
        return df


class CheckColumnOrder(Task):
    """
    Task for checking the order of columns in the loaded DF and in the SQL table into which the data from DF will be loaded.
    If order is different then DF columns are reordered according to the columns of the SQL table.
    """

    def __init__(
        self,
        table: str = None,
        schema: str = None,
        if_exists: Literal["fail", "replace", "append", "delete"] = "replace",
        df: pd.DataFrame = None,
        credentials_secret: str = None,
        vault_name: str = None,
        timeout: int = 3600,
        *args,
        **kwargs,
    ):
        self.credentials_secret = credentials_secret
        self.vault_name = vault_name

        super().__init__(
            name="run_check_column_order", timeout=timeout, *args, **kwargs
        )

    def df_change_order(
        self, df: pd.DataFrame = None, sql_column_list: List[str] = None
    ):
        df_column_list = list(df.columns)
        if set(df_column_list) == set(sql_column_list):
            df_changed = df.loc[:, sql_column_list]
        else:
            raise ValidationError(
                "Detected discrepancies in number of columns or different column names between the CSV file and the SQL table!"
            )

        return df_changed

    def sanitize_columns(self, df: pd.DataFrame = None):
        """
        Function to remove spaces at the end of column name.
        Args:
            df(pd.DataFrame): Dataframe to transform. Defaults to None.
        """
        for col in df.columns:
            df = df.rename(columns={col: col.strip()})
        return df

    def run(
        self,
        table: str = None,
        schema: str = None,
        if_exists: Literal["fail", "replace", "append", "delete"] = "replace",
        df: pd.DataFrame = None,
        credentials_secret: str = None,
        vault_name: str = None,
    ):
        """
        Run a checking column order

        Args:
            table (str, optional): SQL table name without schema. Defaults to None.
            schema (str, optional): SQL schema name. Defaults to None.
            if_exists (Literal, optional): What to do if the table exists. Defaults to "replace".
            df (pd.DataFrame, optional): Data Frame. Defaults to None.
            credentials_secret (str, optional): The name of the Azure Key Vault secret containing a dictionary
            with SQL db credentials (server, db_name, user, and password). Defaults to None.
            vault_name (str, optional): The name of the vault from which to obtain the secret. Defaults to None.
        """
        if if_exists not in ["fail", "replace", "append", "delete"]:
            raise ValueError(
                f"Please select one of the allowed 'if_exists' values: 'fail', 'replace', 'append', 'delete'."
            )
        credentials = get_credentials(credentials_secret, vault_name=vault_name)
        azure_sql = AzureSQL(credentials=credentials)
        df = self.sanitize_columns(df)
        query = f"SELECT COLUMN_NAME FROM INFORMATION_SCHEMA.COLUMNS WHERE TABLE_SCHEMA = '{schema}' AND TABLE_NAME = '{table}'"
        result = azure_sql.run(query=query)
        table_exists = len(result) != 0
        if not table_exists:
            self.logger.warning("Target table doesn't exists.")
            return df
        if if_exists == "fail":
            raise ValueError(
                "The table already exists and 'if_exists' is set to 'fail'."
            )
        if if_exists in ["append", "delete"]:
            sql_column_list = [row[0] for row in result]
            df_column_list = list(df.columns)
            if sql_column_list != df_column_list:
                self.logger.warning(
                    "Detected column order difference between the CSV file and the table. Reordering..."
                )
                df = self.df_change_order(df=df, sql_column_list=sql_column_list)
            else:
                return df
        else:
            self.logger.info("The table will be replaced.")
            return df


class AzureSQLUpsert(Task):
    """Task for upserting data from a pandas DataFrame into AzureSQL.

    Args:
        schema (str, optional): The schema where the data should be upserted. Defaults to None.
        table (str, optional): The table where the data should be upserted. Defaults to None.
        on (str, optional): The field on which to merge (upsert). Defaults to None.
        credentials_secret (str, optional): The name of the Azure Key Vault secret containing a dictionary
        vault_name (str, optional): The name of the vault from which to obtain the secret. Defaults to None.
        timeout(int, optional): The amount of time (in seconds) to wait while running this task before
            a timeout occurs. Defaults to 3600.
    """

    def __init__(
        self,
        schema: str = None,
        table: str = None,
        on: str = None,
        credentials_secret: str = None,
        timeout: int = 3600,
        *args,
        **kwargs,
    ):
        self.schema = schema
        self.table = table
        self.on = on
        self.credentials_secret = credentials_secret
        super().__init__(name="azure_sql_upsert", timeout=timeout, *args, **kwargs)

    @defaults_from_attrs(
        "schema",
        "table",
        "on",
        "credentials_secret",
    )
    def run(
        self,
        df: pd.DataFrame,
        schema: str = None,
        table: str = None,
        on: str = None,
        credentials_secret: str = None,
        vault_name: str = None,
    ):
        """Upsert data from a pandas DataFrame into AzureSQL using a temporary staging table.

        Args:
            df (pd.DataFrame): The DataFrame to upsert.
            schema (str, optional): The schema where the data should be upserted. Defaults to None.
            table (str, optional): The table where the data should be upserted. Defaults to None.
            on (str, optional): The field on which to merge (upsert). Defaults to None.
            credentials_secret (str, optional): The name of the Azure Key Vault secret containing a dictionary
            vault_name (str, optional): The name of the vault from which to obtain the secret. Defaults to None.
        """

        if not table:
            raise ValueError("'table' was not provided.")

        if not on:
            raise ValueError("'on' was not provided.")

        credentials = get_credentials(credentials_secret, vault_name=vault_name)
        azure_sql = AzureSQL(credentials=credentials)

        # Create a temporary staging table.
        # Hashtag marks a temp table in SQL server.
        stg_table = "#" + "stg_" + table
        dtypes = get_sql_server_table_dtypes(
            schema=schema, table=table, con=azure_sql.con
        )
        created = azure_sql.create_table(
            schema=schema, table=stg_table, dtypes=dtypes, if_exists="fail"
        )

        # Insert data into the temp table
        stg_table_fqn = f"{schema}.{stg_table}"
        insert_query = gen_bulk_insert_query_from_df(df, table_fqn=stg_table_fqn)
        inserted = azure_sql.run(insert_query)

        # Upsert into prod table
        merge_query = build_merge_query(
            stg_schema=schema,
            stg_table=stg_table,
            schema=schema,
            table=table,
            primary_key=on,
            con=azure_sql.con,
        )

        merged = azure_sql.run(merge_query)

        if merged:
            rows = df.shape[0]
            table_fqn = f"{schema}.{table}"
            self.logger.info(
                f"Successfully upserted {rows} rows of data into table '{table_fqn}'."
            )

        return True



import inspect
import json
import logging

from prefect.tasks.secrets import PrefectSecret

from viadot.tasks import SQLServerCreateTable, SQLServerQuery, SQLServerToDF
from viadot.tasks.azure_key_vault import AzureKeyVaultSecret

SCHEMA = "sandbox"
TABLE = "test"


def test_sql_server_create_table_init():
    instance = SQLServerCreateTable()
    name = instance.__dict__["name"]
    assert inspect.isclass(SQLServerCreateTable)
    assert isinstance(instance, SQLServerCreateTable)
    assert name == "sql_server_create_table"


def test_sql_server_create_table(caplog):
    credentials_secret = PrefectSecret(
        "AZURE_DEFAULT_SQLDB_SERVICE_PRINCIPAL_SECRET"
    ).run()
    vault_name = PrefectSecret("AZURE_DEFAULT_KEYVAULT").run()
    azure_secret_task = AzureKeyVaultSecret()
    credentials_str = azure_secret_task.run(
        secret=credentials_secret, vault_name=vault_name
    )

    dtypes = {
        "date": "DATE",
        "name": "VARCHAR(255)",
        "id": "VARCHAR(255)",
        "weather": "FLOAT(24)",
        "rain": "FLOAT(24)",
        "temp": "FLOAT(24)",
        "summary": "VARCHAR(255)",
    }

    create_table_task = SQLServerCreateTable()
    with caplog.at_level(logging.INFO):
        create_table_task.run(
            schema=SCHEMA,
            table=TABLE,
            dtypes=dtypes,
            if_exists="replace",
            credentials=json.loads(credentials_str),
        )
        assert "Successfully created table" in caplog.text


def test_sql_server_to_df():
    task = SQLServerToDF(config_key="AZURE_SQL")
    df = task.run(query=f"SELECT * FROM {SCHEMA}.{TABLE}")
    assert df.columns.to_list() == [
        "date",
        "name",
        "id",
        "weather",
        "rain",
        "temp",
        "summary",
    ]


def test_sql_server_query(caplog):
    task = SQLServerQuery(config_key="AZURE_SQL")
    task.run(f"DROP TABLE IF EXISTS sandbox.test_query")
    with caplog.at_level(logging.INFO):
        task.run(f"CREATE TABLE sandbox.test_query (Id INT, Name VARCHAR(10))")
        assert "Successfully ran the query." in caplog.text
    task.run(f"DROP TABLE sandbox.test_query")



import json
from datetime import timedelta
from typing import Literal

from prefect.tasks.secrets import PrefectSecret
from prefect.tasks.shell import ShellTask
from prefect.utilities import logging
from prefect.utilities.tasks import defaults_from_attrs

from .azure_key_vault import AzureKeyVaultSecret

logger = logging.get_logger(__name__)


def parse_logs(log_file_path: str):
    with open(log_file_path) as log_file:
        log_file = log_file.readlines()
    for line in log_file:
        if "#" in line:
            line = line.replace("#", "")
            line = line.replace("@", "")
            logger.warning(line)


class BCPTask(ShellTask):
    """
    Task for bulk inserting data into SQL Server-compatible databases.
    Args:
        - path (str, optional): The path to the local CSV file to be inserted.
        - schema (str, optional): The destination schema.
        - table (str, optional): The destination table.
        - chunksize (int, optional): The chunk size to use.
        - error_log_file_path (string, optional): Full path of an error file. Defaults to "log_file.log".
        - on_error (Literal["skip", "fail"], optional): What to do if error occurs. Defaults to "skip".
        - credentials (dict, optional): The credentials to use for connecting with the database.
        - vault_name (str): The name of the vault from which to fetch the secret.
        - timeout(int, optional): The amount of time (in seconds) to wait while running this task before
            a timeout occurs. Defaults to 3600.
        - **kwargs (dict, optional): Additional keyword arguments to pass to the Task constructor.
    """

    def __init__(
        self,
        path: str = None,
        schema: str = None,
        table: str = None,
        chunksize: int = 5000,
        error_log_file_path: str = "./log_file.log",
        on_error: Literal["skip", "fail"] = "skip",
        credentials: dict = None,
        vault_name: str = None,
        max_retries: int = 3,
        retry_delay: timedelta = timedelta(seconds=10),
        timeout: int = 3600,
        *args,
        **kwargs,
    ):
        self.path = path
        self.schema = schema
        self.table = table
        self.chunksize = chunksize
        self.error_log_file_path = error_log_file_path
        self.on_error = on_error
        self.credentials = credentials
        self.vault_name = vault_name

        super().__init__(
            name="bcp",
            log_stderr=True,
            return_all=True,
            max_retries=max_retries,
            retry_delay=retry_delay,
            timeout=timeout,
            *args,
            **kwargs,
        )

    @defaults_from_attrs(
        "path",
        "schema",
        "table",
        "chunksize",
        "error_log_file_path",
        "on_error",
        "credentials",
        "vault_name",
        "max_retries",
        "retry_delay",
    )
    def run(
        self,
        path: str = None,
        schema: str = None,
        table: str = None,
        chunksize: int = None,
        error_log_file_path: str = None,
        on_error: Literal = None,
        credentials: dict = None,
        credentials_secret: str = None,
        vault_name: str = None,
        max_retries: int = None,
        retry_delay: timedelta = None,
        **kwargs,
    ) -> str:
        """
        Task run method.
        Args:
        - path (str, optional): The path to the local CSV file to be inserted.
        - schema (str, optional): The destination schema.
        - table (str, optional): The destination table.
        - chunksize (int, optional): The chunk size to use. By default 5000.
        - error_log_file_path (string, optional): Full path of an error file. Defaults to "log_file.log".
        - on_error (Literal, optional): What to do if error occur. Defaults to None.
        - credentials (dict, optional): The credentials to use for connecting with SQL Server.
        - credentials_secret (str, optional): The name of the Key Vault secret containing database credentials.
        (server, db_name, user, password)
        - vault_name (str): The name of the vault from which to fetch the secret.
        Returns:
            str: The output of the bcp CLI command.
        """
        if not credentials:
            if not credentials_secret:
                # attempt to read a default for the service principal secret name
                try:
                    credentials_secret = PrefectSecret(
                        "AZURE_DEFAULT_SQLDB_SERVICE_PRINCIPAL_SECRET"
                    ).run()
                except ValueError:
                    pass

            if credentials_secret:
                credentials_str = AzureKeyVaultSecret(
                    credentials_secret, vault_name=vault_name
                ).run()
                credentials = json.loads(credentials_str)

        fqn = f"{schema}.{table}" if schema else table

        server = credentials["server"]
        db_name = credentials["db_name"]
        uid = credentials["user"]
        pwd = credentials["password"]

        if "," in server:
            # A space after the comma is allowed in the ODBC connection string
            # but not in BCP's 'server' argument.
            server = server.replace(" ", "")

        if on_error == "skip":
            max_error = 0
        elif on_error == "fail":
            max_error = 1
        else:
            raise ValueError(
                "Please provide correct 'on_error' parameter value - 'skip' or 'fail'. "
            )
        command = f"/opt/mssql-tools/bin/bcp {fqn} in '{path}' -S {server} -d {db_name} -U {uid} -P '{pwd}' -c -F 2 -b {chunksize} -h 'TABLOCK' -e '{error_log_file_path}' -m {max_error}"
        run_command = super().run(command=command, **kwargs)
        try:
            parse_logs(error_log_file_path)
        except:
            logger.warning("BCP logs couldn't be parsed.")
        return run_command



import json
from typing import Any, Dict, List

from viadot.exceptions import CredentialError
from viadot.sources.base import Source
from viadot.utils import handle_api_response


class Hubspot(Source):
    """
    A class that connects and extracts data from Hubspot API. Documentation is available under https://developers.hubspot.com/docs/api/crm/understanding-the-crm.
    Connector allows to pull data in three ways:
        - using base API for crm schemas as an endpoint (eg. "contacts", ""line_items", "deals", ...),
        - using full url as endpoint.
    """

    def __init__(self, credentials: dict, *args, **kwargs):
        """
        Create an instance of Hubspot.

        Args:
            credentials (dict): Credentials to Hubspot API.
        """

        self.credentials = credentials
        if credentials is not None:
            try:
                self.headers = {
                    "Authorization": f'Bearer {self.credentials["TOKEN"]}',
                    "Content-Type": "application/json",
                }
            except:
                raise CredentialError("Credentials not found.")

        super().__init__(*args, credentials=self.credentials, **kwargs)

        self.base_url = self.credentials.get("URL")

    def clean_special_characters(self, value: str = None) -> str:
        """
        Function that maps special characters from ASCII to HTML5 equivalents.

        Args:
            value (str, optional): Value for decoding. Defaults to None.

        Returns:
            str: Value after decoding to HTML5.
        """

        mapping = {
            "!": "%21",
            '"': "%22",
            "#": "%23",
            "$": "%24",
            "%": "%25",
            "&": "%26",
            "(": "%28",
            ")": "%29",
            "*": "%2A",
            "+": "%2B",
            ",": "%2C",
        }

        for key in mapping.keys():
            if key in value:
                value = value.replace(key, mapping[key])

        return value

    def get_properties_url(self, endpoint: str = None) -> str:
        """
        Function that generates URL with all properties ( = detailed information about all avaliable columns for defined endpoint (schema)).

        Args:
            endpoint (str, optional): Hubspot schema from crm objects (eg. contacts, line_items, ...). Defaults to None.

        Returns:
            str: Url with properties for defined endpoint.
        """

        url = f"{self.base_url}/properties/v2/{endpoint}/properties"
        return url

    def get_api_url(
        self,
        endpoint: str = None,
        filters: Dict[str, Any] = None,
        properties: List[Any] = None,
    ) -> str:
        """
        Function that generates full url for Hubspot API with defined parametrs.

        Args:
            endpoint (str, optional): Schema or full url. Defaults to None.
            filters (Dict[str, Any], optional): Filters defined for the API body in specific order. Defaults to None.
            properties (List[Any], optional): List of user-defined columns to be pulled from the API. Defaults to None.

        Returns:
            str: Generated url passed to Hubspot API.
        """

        if self.base_url in endpoint:
            url = endpoint
        else:
            if endpoint.startswith("hubdb"):
                url = f"{self.base_url}/{endpoint}"
            else:
                if filters:
                    url = (
                        f"{self.base_url}/crm/v3/objects/{endpoint}/search/?limit=100&"
                    )
                else:
                    url = f"{self.base_url}/crm/v3/objects/{endpoint}/?limit=100&"

                if properties and len(properties) > 0:
                    url += f'properties={",".join(properties)}&'

        return url

    def get_api_body(self, filters: Dict[str, Any] = {}) -> Dict:
        """
        Function that cleans the filters body and converts to a JSON formatted value.

        Args:
            filters (Dict[str, Any], optional): Filters dictionary that will be passed to Hubspot API. Defaults to {}.

                    Example below:

                        filters = {
                                "filters": [
                                    {
                                    "propertyName": "createdate",
                                    "operator": "BETWEEN",
                                    "highValue": "2023-03-27",
                                    "value": "2023-03-26"
                                    }
                                ]
                                }

                    Operators between the min and max value are listed below:
                    [IN, NOT_HAS_PROPERTY, LT, EQ, GT, NOT_IN, GTE, CONTAINS_TOKEN, HAS_PROPERTY, LTE, NOT_CONTAINS_TOKEN, BETWEEN, NEQ]

                    LT - Less than
                    LTE - Less than or equal to
                    GT - Greater than
                    GTE - Greater than or equal to
                    EQ - Equal to
                    NEQ - Not equal to
                    BETWEEN - Within the specified range. In your request, use key-value pairs to set highValue and value. Refer to the example above.
                    IN - Included within the specified list. This operator is case-sensitive, so inputted values must be in lowercase.
                    NOT_IN - Not included within the specified list
                    HAS_PROPERTY - Has a value for the specified property
                    NOT_HAS_PROPERTY - Doesn't have a value for the specified property
                    CONTAINS_TOKEN - Contains a token. In your request, you can use wildcards (*) to complete a partial search. For example, use the value *@hubspot.com to retrieve contacts with a HubSpot email address.
                    NOT_CONTAINS_TOKEN  -Doesn't contain a token

        Returns:
            Dict: Filters with a JSON format.
        """
        payload = json.dumps({"filterGroups": filters, "limit": 100})

        return payload

    def to_json(self, url: str = None, body: str = None, method: str = None) -> Dict:
        """
        Function that converts API response to a JSON formatted data.

        Args:
            url (str, optional): Hubspot API url. Defaults to None.
            body (str, optional): Filters that will be pushed to the API body. Defaults to None.
            method (str, optional): Method of the API call ("GET"/"POST"). Defaults to None.

        Returns:
            Dict: API response in JSON format.
        """

        response = handle_api_response(
            url=url, headers=self.headers, body=body, method=method
        )

        return response.json()



import os
from datetime import datetime
from unittest import mock

import pandas as pd
import pytest
from O365.account import Account

from viadot.exceptions import CredentialError
from viadot.sources import Outlook


@pytest.fixture
def var_dictionary():
    variables = {
        "mailbox_name": "random@random.com",
        "start_date": "2023-04-12",
        "end_date": "2023-04-13",
        "credentials": {
            "client_id": "abcdefghijklmnopqrstuvwxyz",
            "client_secret": "abcdefghijklmnopqrstuvwxyz",
            "tenant_id": "abcdefghijklmnopqrstuvwxyz",
        },
        "final_dict_folders": {"Mailbox": MockValue},
    }

    return variables


class MockClass:
    mailbox_name = "Trial and error"

    def authenticate():
        return True

    def mailbox():
        return None


class MockValue:
    name = "mailbox"

    def get_messages(limit=1):
        return [MockMessage]


class MockMessage:
    subject = "subject"
    received = "2023-04-12T06:09:59+00:00"
    categories = ["categories"]
    conversation_index = "aaaaaaaaaaaaaaaaaaaaaaaaaaaa"

    def to_api_data():
        data = {
            "toRecipients": [
                {
                    "emailAddress": {
                        "address": "random@random.com",
                        "name": "random",
                    }
                },
                {
                    "emailAddress": {
                        "address": "random@random2.com",
                        "name": "random",
                    }
                },
            ],
            "from": {"emailAddress": {"address": "random@random.ee", "name": "name"}},
            "receivedDateTime": "2022-04-01T06:09:59+00:00",
            "conversationId": "bbbb",
        }
        return data


@pytest.mark.basics
def test_outlook_credentials(var_dictionary):
    o = Outlook(
        mailbox_name=var_dictionary["mailbox_name"],
        credentials=var_dictionary["credentials"],
    )

    assert all(
        [
            isinstance(o.credentials, dict),
            isinstance(o.date_range_end_time, datetime),
            isinstance(o.date_range_start_time, datetime),
            isinstance(o.account, Account),
        ]
    )


@pytest.mark.basics
@mock.patch("viadot.sources.outlook.Account", return_value=MockClass)
def test_outlook_mailbox_limit(mock_api_Account, var_dictionary):
    o = Outlook(
        mailbox_name=var_dictionary["mailbox_name"],
        credentials=var_dictionary["credentials"],
        start_date=var_dictionary["start_date"],
        end_date=var_dictionary["end_date"],
    )
    data = o._get_messages_from_mailbox(
        var_dictionary["final_dict_folders"], address_limit=20
    )
    assert len(data[0]["recivers"]) <= 20


@pytest.mark.basics
@mock.patch("viadot.sources.outlook.Account", return_value=MockClass)
def test_outlook_mailbox_space(mock_api_Account, var_dictionary):
    o = Outlook(
        mailbox_name=var_dictionary["mailbox_name"],
        credentials=var_dictionary["credentials"],
        start_date=var_dictionary["start_date"],
        end_date=var_dictionary["end_date"],
    )
    data = o._get_messages_from_mailbox(
        var_dictionary["final_dict_folders"], address_limit=5
    )
    assert len(data[0]["recivers"]) == 0


@pytest.mark.exceptions
def test_outlook_credential_exception(var_dictionary):
    with pytest.raises(CredentialError):
        o = Outlook(
            mailbox_name=var_dictionary["mailbox_name"],
            start_date=var_dictionary["start_date"],
            end_date=var_dictionary["end_date"],
        )


@pytest.mark.exceptions
def test_outlook_mailbox_folders_exception(var_dictionary):
    with pytest.raises(Exception):
        o = Outlook(
            mailbox_name=var_dictionary["mailbox_name"],
            start_date=var_dictionary["start_date"],
            end_date=var_dictionary["end_date"],
            mailbox_folders=["trial"],
        )


@mock.patch.object(Outlook, "get_all_mails_to_df", return_value=pd.DataFrame())
@mock.patch("viadot.sources.outlook.Account", return_value=MockClass)
@pytest.mark.to_csv
def test_outlook_to_csv(mock_method_Outlook, mock_api_Account, var_dictionary):
    o = Outlook(
        mailbox_name=var_dictionary["mailbox_name"],
        start_date=var_dictionary["start_date"],
        end_date=var_dictionary["end_date"],
        credentials=var_dictionary["credentials"],
    )
    df = o.get_all_mails_to_df()
    o.to_csv(df)
    file_name = (
        var_dictionary["mailbox_name"].split("@")[0].replace(".", "_").replace("-", "_")
    )
    mock_method_Outlook.assert_called()
    mock_api_Account.assert_called()
    assert os.path.exists(f"{file_name}.csv")



import os

from viadot.config import local_config
from viadot.flows import EpicorOrdersToDuckDB
from viadot.tasks import DuckDBQuery, DuckDBToDF

TABLE = "test_epicor"
SCHEMA = "sandbox"
LOCAL_PATH = "test_epicor.parquet"


def test_epicor_to_duckdb():
    duckdb_creds = {"database": "/home/viadot/database/test.duckdb"}
    flow = EpicorOrdersToDuckDB(
        name="test",
        epicor_config_key="EPICOR",
        base_url=local_config.get("EPICOR").get("test_url"),
        filters_xml="""
    <OrderQuery>
        <QueryFields>
            <CompanyNumber>001</CompanyNumber>
            <BegInvoiceDate>2022-05-16</BegInvoiceDate>
            <EndInvoiceDate>2022-05-16</EndInvoiceDate>
            <RecordCount>3</RecordCount>
        </QueryFields>
    </OrderQuery>""",
        if_exists="replace",
        duckdb_table=TABLE,
        duckdb_schema=SCHEMA,
        duckdb_credentials=duckdb_creds,
        local_file_path=LOCAL_PATH,
    )

    result = flow.run()
    assert result.is_successful()

    df_task = DuckDBToDF(credentials=duckdb_creds)
    df = df_task.run(table=TABLE, schema=SCHEMA)

    assert df.shape == (24, 59)

    run_query = DuckDBQuery()
    run_query.run(query=f"DROP TABLE {SCHEMA}.{TABLE}", credentials=duckdb_creds)
    os.remove(LOCAL_PATH)



import copy
import json
import os
from datetime import timedelta
from typing import Any, Dict, List, Literal

import pandas as pd
from prefect import Task
from prefect.tasks.secrets import PrefectSecret
from prefect.utilities import logging
from prefect.utilities.tasks import defaults_from_attrs

from viadot.task_utils import *

from ..sources import VidClub

logger = logging.get_logger()


class VidClubToDF(Task):
    def __init__(
        self,
        source: Literal["jobs", "product", "company", "survey"] = None,
        credentials: Dict[str, Any] = None,
        credentials_secret: str = "VIDCLUB",
        vault_name: str = None,
        from_date: str = "2022-03-22",
        to_date: str = None,
        timeout: int = 3600,
        report_name: str = "vid_club_to_df",
        *args: List[Any],
        **kwargs: Dict[str, Any],
    ):
        """
        Task to downloading data from Vid Club APIs to Pandas DataFrame.

        Args:
            source (Literal["jobs", "product", "company", "survey"], optional): The endpoint source to be accessed. Defaults to None.
            credentials (Dict[str, Any], optional): Stores the credentials information. Defaults to None.
            credentials_secret (str, optional): The name of the secret in Azure Key Vault or Prefect or local_config file. Defaults to "VIDCLUB".
            vault_name (str, optional): For credentials stored in Azure Key Vault. The name of the vault from which to obtain the secret. Defaults to None.
            from_date (str): Start date for the query, by default is the oldest date in the data, '2022-03-22'.
            to_date (str, optional): End date for the query. By default None, which will be executed as datetime.today().strftime("%Y-%m-%d") in code.
            timeout (int, optional): The amount of time (in seconds) to wait while running this task before
                a timeout occurs. Defaults to 3600.
            report_name (str, optional): Stores the report name. Defaults to "vid_club_to_df".

        Returns: Pandas DataFrame
        """
        self.source = source
        self.from_date = from_date
        self.to_date = to_date
        self.report_name = report_name
        self.credentials_secret = credentials_secret
        self.vault_name = vault_name

        if credentials is None:
            self.credentials = credentials_loader.run(
                credentials_secret=credentials_secret, vault_name=vault_name
            )
        else:
            self.credentials = credentials

        super().__init__(
            name=self.report_name,
            timeout=timeout,
            *args,
            **kwargs,
        )

    def __call__(self, *args, **kwargs):
        """Download Vid Club data to Pandas DataFrame"""
        return super().__call__(*args, **kwargs)

    @defaults_from_attrs(
        "source",
        "credentials",
        "from_date",
        "to_date",
    )
    def run(
        self,
        source: Literal["jobs", "product", "company", "survey"] = None,
        credentials: Dict[str, Any] = None,
        from_date: str = "2022-03-22",
        to_date: str = None,
        items_per_page: int = 100,
        region: str = "all",
        days_interval: int = 30,
        cols_to_drop: List[str] = None,
    ) -> pd.DataFrame:
        """
        Task run method.

        Args:
            source (Literal["jobs", "product", "company", "survey"], optional): The endpoint source to be accessed. Defaults to None.
            credentials (Dict[str, Any], optional): Stores the credentials information. Defaults to None.
            from_date (str, optional): Start date for the query, by default is the oldest date in the data, '2022-03-22'.
            to_date (str, optional): End date for the query. By default None, which will be executed as datetime.today().strftime("%Y-%m-%d") in code.
            items_per_page (int, optional): Number of entries per page. 100 entries by default.
            region (str, optional): Region filter for the query. Valid inputs: ["bg", "hu", "hr", "pl", "ro", "si", "all"]. Defaults to "all".
            days_interval (int, optional): Days specified in date range per api call (test showed that 30-40 is optimal for performance). Defaults to 30.
            cols_to_drop (List[str], optional): List of columns to drop. Defaults to None.

        Raises:
            KeyError: When DataFrame doesn't contain columns provided in the list of columns to drop.
            TypeError: When cols_to_drop is not a list type.

        Returns:
            pd.DataFrame: The query result as a pandas DataFrame.
        """

        vc_obj = VidClub(credentials=credentials)

        vc_dataframe = vc_obj.total_load(
            source=source,
            from_date=from_date,
            to_date=to_date,
            items_per_page=items_per_page,
            region=region,
            days_interval=days_interval,
        )
        if cols_to_drop is not None:
            if isinstance(cols_to_drop, list):
                try:
                    logger.info(f"Dropping following columns: {cols_to_drop}...")
                    vc_dataframe.drop(
                        columns=cols_to_drop, inplace=True, errors="raise"
                    )
                except KeyError as ke:
                    logger.error(
                        f"Column(s): {cols_to_drop} don't exist in the DataFrame. No columns were dropped. Returning full DataFrame..."
                    )
                    logger.info(f"Existing columns: {vc_dataframe.columns}")
            else:
                raise TypeError("Provide columns to drop in a List.")

        return vc_dataframe



from typing import Literal

from prefect.utilities import logging

from .sql_server import SQLServer

logger = logging.get_logger(__name__)


class AzureSQL(SQLServer):
    def __init__(self, *args, config_key="AZURE_SQL", **kwargs):
        super().__init__(*args, config_key=config_key, **kwargs)

    def bulk_insert(
        self,
        table: str,
        schema: str = None,
        source_path: str = None,
        sep="\t",
        if_exists: Literal = "append",
    ):
        """Fuction to bulk insert.
        Args:
            table (str): Table name.
            schema (str, optional): Schema name. Defaults to None.
            source_path (str, optional): Full path to a data file. Defaults to one.
            sep (str, optional):  field terminator to be used for char and widechar data files. Defaults to "\t".
            if_exists (Literal, optional): What to do if the table already exists. Defaults to "append".
        """
        if schema is None:
            schema = self.DEFAULT_SCHEMA
        fqn = f"{schema}.{table}"
        insert_sql = f"""
            BULK INSERT {fqn} FROM '{source_path}'
            WITH (
                CHECK_CONSTRAINTS,
                DATA_SOURCE='{self.credentials['data_source']}',
                DATAFILETYPE='char',
                FIELDTERMINATOR='{sep}',
                ROWTERMINATOR='0x0a',
                FIRSTROW=2,
                KEEPIDENTITY,
                TABLOCK,
                CODEPAGE='65001'
            );
        """
        if if_exists == "replace":
            self.run(f"DELETE FROM {schema}.{table}")
        self.run(insert_sql)
        return True

    def create_external_database(
        self,
        external_database_name: str,
        storage_account_name: str,
        container_name: str,
        sas_token: str,
        master_key_password: str,
        credential_name: str = None,
    ):
        """Create an external database. Used to eg. execute BULK INSERT or OPENROWSET
        queries.

        Args:
            external_database_name (str): The name of the extrnal source (db) to be created.
            storage_account_name (str): The name of the Azure storage account.
            container_name (str): The name of the container which should become the "database".
            sas_token (str): The SAS token to be used as the credential. Note that the auth
            system in Azure is pretty broken and you might need to paste here your storage
            account's account key instead.
            master_key_password (str): The password for the database master key of your
            Azure SQL Database.
            credential_name (str): How to name the SAS credential. This is really an Azure
            internal thing and can be anything. By default '{external_database_name}_credential`.
        """

        # stupid Microsoft thing
        if sas_token.startswith("?"):
            sas_token = sas_token[1:]

        if credential_name is None:
            credential_name = f"{external_database_name}_credential"

        create_master_key_sql = (
            f"CREATE MASTER KEY ENCRYPTION BY PASSWORD = {master_key_password}"
        )

        create_external_db_credential_sql = f"""
        CREATE DATABASE SCOPED CREDENTIAL {credential_name}
        WITH IDENTITY = 'SHARED ACCESS SIGNATURE'
        SECRET = '{sas_token}';
        """

        create_external_db_sql = f"""
        CREATE EXTERNAL DATA SOURCE {external_database_name} WITH (
        LOCATION = f'https://{storage_account_name}.blob.core.windows.net/{container_name}',
        CREDENTIAL = {credential_name}
        );
        """

        self.run(create_master_key_sql)
        self.run(create_external_db_credential_sql)
        self.run(create_external_db_sql)



from typing import Any, Dict, List

from prefect import Flow, task
from prefect.storage import Local
from prefect.utilities import logging

from ..tasks import AzureDataLakeCopy

logger = logging.get_logger(__name__)


@task(timeout=3600)
def is_stored_locally(f: Flow):
    return f.storage is None or isinstance(f.storage, Local)


class ADLSContainerToContainer(Flow):
    """Copy file(s) between containers.

    Args:
        name (str): The name of the flow.
        from_path (str): The path to the Data Lake file.
        to_path (str): The path of the final file location a/a/filename.extension.
        adls_sp_credentials_secret (str, optional): The name of the Azure Key Vault secret containing a dictionary with
            ACCOUNT_NAME and Service Principal credentials (TENANT_ID, CLIENT_ID, CLIENT_SECRET) for the Azure Data Lake.
            Defaults to None.
        vault_name (str): The name of the vault from which to retrieve the secrets.
        timeout(int, optional): The amount of time (in seconds) to wait while running this task before
            a timeout occurs. Defaults to 3600.
    """

    def __init__(
        self,
        name: str,
        from_path: str,
        to_path: str,
        adls_sp_credentials_secret: str = None,
        vault_name: str = None,
        timeout: int = 3600,
        *args: List[any],
        **kwargs: Dict[str, Any]
    ):
        from_path = from_path.strip("/")
        self.from_path = from_path
        self.to_path = to_path
        self.adls_sp_credentials_secret = adls_sp_credentials_secret
        self.vault_name = vault_name
        self.timeout = timeout
        super().__init__(*args, name=name, **kwargs)
        self.gen_flow()

    @staticmethod
    def slugify(name):
        return name.replace(" ", "_").lower()

    def gen_flow(self) -> Flow:
        copy_task = AzureDataLakeCopy(timeout=self.timeout)
        copy_task.bind(
            from_path=self.from_path,
            to_path=self.to_path,
            sp_credentials_secret=self.adls_sp_credentials_secret,
            vault_name=self.vault_name,
            flow=self,
        )



import json
import os
from typing import Any, Dict, List, Literal

import pandas as pd
from prefect import Flow, task
from prefect.backend import get_key_value
from prefect.engine import signals
from prefect.utilities import logging

from viadot.tasks import (
    AzureDataLakeCopy,
    AzureDataLakeToDF,
    AzureSQLCreateTable,
    AzureSQLDBQuery,
    BCPTask,
    CheckColumnOrder,
    DownloadGitHubFile,
)
from viadot.tasks.azure_data_lake import AzureDataLakeDownload

logger = logging.get_logger(__name__)


@task(timeout=3600)
def union_dfs_task(dfs: List[pd.DataFrame]):
    return pd.concat(dfs, ignore_index=True)


@task(timeout=3600)
def map_data_types_task(json_shema_path: str):
    file_dtypes = open(json_shema_path)
    dict_dtypes = json.load(file_dtypes)
    dict_mapping = {
        "Float": "REAL",
        "Image": None,
        "Categorical": "VARCHAR(500)",
        "Time": "TIME",
        "Boolean": "BIT",
        "DateTime": "DATETIMEOFFSET",  # DATETIMEOFFSET is the only timezone-aware dtype in TSQL
        "Object": "VARCHAR(500)",
        "EmailAddress": "VARCHAR(50)",
        "File": None,
        "Geometry": "GEOMETRY",
        "Ordinal": "VARCHAR(500)",
        "Integer": "INT",
        "Generic": "VARCHAR(500)",
        "UUID": "UNIQUEIDENTIFIER",
        "Complex": None,
        "Date": "DATE",
        "String": "VARCHAR(500)",
        "IPAddress": "VARCHAR(39)",
        "Path": "VARCHAR(500)",
        "TimeDelta": "VARCHAR(20)",  # datetime.datetime.timedelta; eg. '1 days 11:00:00'
        "URL": "VARCHAR(500)",
        "Count": "INT",
    }
    dict_dtypes_mapped = {}
    for k in dict_dtypes:
        dict_dtypes_mapped[k] = dict_mapping[dict_dtypes[k]]
    return dict_dtypes_mapped


@task(timeout=3600)
def df_to_csv_task(df, remove_tab, path: str, sep: str = "\t"):
    # if table doesn't exist it will be created later -  df equals None
    if df is None:
        logger.warning("DataFrame is None")
    else:
        if remove_tab == True:
            for col in range(len(df.columns)):
                df[df.columns[col]] = (
                    df[df.columns[col]].astype(str).str.replace(r"\t", "", regex=True)
                )
            df.to_csv(path, sep=sep, index=False)
        else:
            df.to_csv(path, sep=sep, index=False)


@task(timeout=3600)
def check_dtypes_sort(
    df: pd.DataFrame = None,
    dtypes: Dict[str, Any] = None,
    apply: bool = True,
) -> Dict[str, Any]:
    """Check dtype column order to avoid malformation SQL table.
    When data is loaded by the user, a data frame is passed to this task
    to check the column sort with dtypes and re-sort if neccessary.
    Args:
        df (pd.DataFrame, optional): Data Frame from original ADLS file. Defaults to None.
        dtypes (Dict[str, Any], optional): Dictionary of columns and data type to apply
            to the Data Frame downloaded. Defaults to None.
        apply (bool, optiona): By default, this task will be used by all the flows. It can
            be set up to False for avoiding its application. Defaults to True.
    Returns:
        Dict[str, Any]: Sorted dtype.
    """
    if df is None:
        logger.error("DataFrame argument is mandatory")
        raise signals.FAIL("DataFrame is None.")
    elif apply:
        # first check if all dtypes keys are in df.columns
        if all(d in df.columns for d in list(dtypes.keys())) and len(df.columns) == len(
            list(dtypes.keys())
        ):
            # check if have the same sort
            matches = list(map(lambda x, y: x == y, df.columns, dtypes.keys()))
            if not all(matches):
                logger.warning(
                    "Some keys are not sorted in dtypes. Repositioning the key:value..."
                )
                # re-sort in a new dtype
                new_dtypes = dict()
                for key in df.columns:
                    new_dtypes.update([(key, dtypes[key])])
            else:
                new_dtypes = dtypes.copy()
        else:
            logger.error("There is a discrepancy with any of the columns.")
            raise signals.FAIL(
                "dtype dictionary contains key(s) that not matching with the ADLS file columns name, or they have different length."
            )
    else:
        new_dtypes = dtypes.copy()

    return new_dtypes


class ADLSToAzureSQL(Flow):
    def __init__(
        self,
        name: str,
        local_file_path: str = None,
        adls_path: str = None,
        read_sep: str = "\t",
        write_sep: str = "\t",
        remove_tab: bool = False,
        overwrite_adls: bool = True,
        if_empty: str = "warn",
        adls_sp_credentials_secret: str = None,
        dtypes: Dict[str, Any] = None,
        check_dtypes_order: bool = True,
        table: str = None,
        schema: str = None,
        if_exists: Literal["fail", "replace", "append", "delete"] = "replace",
        check_col_order: bool = True,
        sqldb_credentials_secret: str = None,
        on_bcp_error: Literal["skip", "fail"] = "skip",
        max_download_retries: int = 5,
        tags: List[str] = ["promotion"],
        vault_name: str = None,
        timeout: int = 3600,
        *args: List[any],
        **kwargs: Dict[str, Any],
    ):
        """
        Flow for downloading data from different marketing APIs to a local CSV
        using Supermetrics API, then uploading it to Azure Data Lake,
        and finally inserting into Azure SQL Database.
        Args:
            name (str): The name of the flow.
            local_file_path (str, optional): Local destination path. Defaults to None.
            adls_path (str): The path to an ADLS folder or file. If you pass a path to a directory,
            the latest file from that directory will be loaded. We assume that the files are named using timestamps.
            read_sep (str, optional): The delimiter for the source file. Defaults to "\t".
            write_sep (str, optional): The delimiter for the output CSV file. Defaults to "\t".
            remove_tab (bool, optional): Whether to remove tab delimiters from the data. Defaults to False.
            overwrite_adls (bool, optional): Whether to overwrite the file in ADLS. Defaults to True.
            if_empty (str, optional): What to do if the Supermetrics query returns no data. Defaults to "warn".
            adls_sp_credentials_secret (str, optional): The name of the Azure Key Vault secret containing a dictionary with
            ACCOUNT_NAME and Service Principal credentials (TENANT_ID, CLIENT_ID, CLIENT_SECRET) for the Azure Data Lake.
            Defaults to None.
            dtypes (dict, optional): Which custom data types should be used for SQL table creation task.
            To be used only in case that dtypes need to be manually mapped - dtypes from raw schema file in use by default. Defaults to None.
            check_dtypes_order (bool, optiona): By default, this task will be used by all the flows. It can
                be set up to False for avoiding its application. Defaults to True.
            table (str, optional): Destination table. Defaults to None.
            schema (str, optional): Destination schema. Defaults to None.
            if_exists (Literal, optional): What to do if the table exists. Defaults to "replace".
            check_col_order (bool, optional): Whether to check column order. Defaults to True.
            sqldb_credentials_secret (str, optional): The name of the Azure Key Vault secret containing a dictionary with
            Azure SQL Database credentials. Defaults to None.
            on_bcp_error (Literal["skip", "fail"], optional): What to do if error occurs. Defaults to "skip".
            max_download_retries (int, optional): How many times to retry the download. Defaults to 5.
            tags (List[str], optional): Flow tags to use, eg. to control flow concurrency. Defaults to ["promotion"].
            vault_name (str, optional): The name of the vault from which to obtain the secrets. Defaults to None.
            timeout(int, optional): The amount of time (in seconds) to wait while running this task before
                a timeout occurs. Defaults to 3600.
        """

        adls_path = adls_path.strip("/")
        # Read parquet
        if adls_path.split(".")[-1] in ["csv", "parquet"]:
            self.adls_path = adls_path
        else:
            self.adls_path = get_key_value(key=adls_path)

        # Read schema
        self.dtypes = dtypes
        self.check_dtypes_order = check_dtypes_order
        self.adls_root_dir_path = os.path.split(self.adls_path)[0]
        self.adls_file_name = os.path.split(self.adls_path)[-1]
        extension = os.path.splitext(self.adls_path)[-1]
        json_schema_file_name = self.adls_file_name.replace(extension, ".json")
        self.json_shema_path = os.path.join(
            self.adls_root_dir_path,
            "schema",
            json_schema_file_name,
        )

        # AzureDataLakeUpload
        self.local_file_path = local_file_path or self.slugify(name) + ".csv"
        self.local_json_path = self.slugify(name) + ".json"
        self.read_sep = read_sep
        self.write_sep = write_sep
        self.overwrite_adls = overwrite_adls
        self.if_empty = if_empty
        self.adls_sp_credentials_secret = adls_sp_credentials_secret
        self.adls_path_conformed = self.get_promoted_path(env="conformed")
        self.adls_path_operations = self.get_promoted_path(env="operations")

        # AzureSQLCreateTable
        self.table = table
        self.schema = schema
        self.if_exists = if_exists
        self.check_col_order = check_col_order
        # Generate CSV
        self.remove_tab = remove_tab

        # BCPTask
        self.sqldb_credentials_secret = sqldb_credentials_secret
        self.on_bcp_error = on_bcp_error

        # Global
        self.max_download_retries = max_download_retries
        self.tags = tags
        self.vault_name = vault_name
        self.timeout = timeout

        super().__init__(*args, name=name, **kwargs)

        # self.dtypes.update(METADATA_COLUMNS)
        self.gen_flow()

    @staticmethod
    def _map_if_exists(if_exists: str) -> str:
        mapping = {"append": "skip"}
        return mapping.get(if_exists, if_exists)

    @staticmethod
    def slugify(name):
        return name.replace(" ", "_").lower()

    def get_promoted_path(self, env: str) -> str:
        adls_path_clean = self.adls_path.strip("/")
        extension = adls_path_clean.split(".")[-1].strip()
        if extension == "parquet":
            file_name = adls_path_clean.split("/")[-2] + ".csv"
            common_path = "/".join(adls_path_clean.split("/")[1:-2])
        else:
            file_name = adls_path_clean.split("/")[-1]
            common_path = "/".join(adls_path_clean.split("/")[1:-1])

        promoted_path = os.path.join(env, common_path, file_name)

        return promoted_path

    def gen_flow(self) -> Flow:
        lake_to_df_task = AzureDataLakeToDF(timeout=self.timeout)
        df = lake_to_df_task.bind(
            path=self.adls_path,
            sp_credentials_secret=self.adls_sp_credentials_secret,
            sep=self.read_sep,
            flow=self,
        )

        if not self.dtypes:
            download_json_file_task = AzureDataLakeDownload(timeout=self.timeout)
            download_json_file_task.bind(
                from_path=self.json_shema_path,
                to_path=self.local_json_path,
                sp_credentials_secret=self.adls_sp_credentials_secret,
                flow=self,
            )
            dtypes = map_data_types_task.bind(self.local_json_path, flow=self)
            map_data_types_task.set_upstream(download_json_file_task, flow=self)
        else:
            dtypes = check_dtypes_sort.bind(
                df,
                dtypes=self.dtypes,
                apply=self.check_dtypes_order,
                flow=self,
            )

        check_column_order_task = CheckColumnOrder(timeout=self.timeout)
        df_reorder = check_column_order_task.bind(
            table=self.table,
            schema=self.schema,
            df=df,
            if_exists=self.if_exists,
            credentials_secret=self.sqldb_credentials_secret,
            flow=self,
        )
        if self.check_col_order == False:
            df_to_csv = df_to_csv_task.bind(
                df=df,
                path=self.local_file_path,
                sep=self.write_sep,
                remove_tab=self.remove_tab,
                flow=self,
            )
        else:
            df_to_csv = df_to_csv_task.bind(
                df=df_reorder,
                path=self.local_file_path,
                sep=self.write_sep,
                remove_tab=self.remove_tab,
                flow=self,
            )

        promote_to_conformed_task = AzureDataLakeCopy(timeout=self.timeout)
        promote_to_conformed_task.bind(
            from_path=self.adls_path,
            to_path=self.adls_path_conformed,
            sp_credentials_secret=self.adls_sp_credentials_secret,
            vault_name=self.vault_name,
            flow=self,
        )
        promote_to_operations_task = AzureDataLakeCopy(timeout=self.timeout)
        promote_to_operations_task.bind(
            from_path=self.adls_path_conformed,
            to_path=self.adls_path_operations,
            sp_credentials_secret=self.adls_sp_credentials_secret,
            vault_name=self.vault_name,
            flow=self,
        )
        create_table_task = AzureSQLCreateTable(timeout=self.timeout)
        create_table_task.bind(
            schema=self.schema,
            table=self.table,
            dtypes=dtypes,
            if_exists=self._map_if_exists(self.if_exists),
            credentials_secret=self.sqldb_credentials_secret,
            vault_name=self.vault_name,
            flow=self,
        )
        bulk_insert_task = BCPTask(timeout=self.timeout)
        bulk_insert_task.bind(
            path=self.local_file_path,
            schema=self.schema,
            table=self.table,
            error_log_file_path=self.name.replace(" ", "_") + ".log",
            on_error=self.on_bcp_error,
            credentials_secret=self.sqldb_credentials_secret,
            vault_name=self.vault_name,
            flow=self,
        )

        df_reorder.set_upstream(lake_to_df_task, flow=self)
        df_to_csv.set_upstream(df_reorder, flow=self)
        promote_to_conformed_task.set_upstream(df_to_csv, flow=self)
        create_table_task.set_upstream(df_to_csv, flow=self)
        promote_to_operations_task.set_upstream(promote_to_conformed_task, flow=self)
        bulk_insert_task.set_upstream(create_table_task, flow=self)



from datetime import datetime, timezone
from typing import Any, Dict, List

import pandas as pd
from prefect import Flow, task
from prefect.utilities import logging

from ..tasks import AzureDataLakeDownload, AzureDataLakeUpload

logger = logging.get_logger(__name__)


@task
def add_ingestion_metadata(
    path: str,
    sep: str = "\t",
):
    """Add ingestion metadata column(s), eg. data download date"""
    df = pd.read_csv(path, sep=sep)
    df["_viadot_downloaded_at_utc"] = datetime.now(timezone.utc)
    df.to_csv(path, sep=sep, index=False)


class ADLSGen1ToGen2(Flow):
    """Move file(s) from Azure Data Lake gen1 to gen2.

    Args:
        name (str): The name of the flow.
        gen1_path (str): The path to the gen1 Data Lake file/folder.
        gen2_path (str): The path of the final gen2 file/folder.
        local_file_path (str): Where the gen1 file should be downloaded.
        overwrite (str): Whether to overwrite the destination file(s).
        gen1_sp_credentials_secret (str): The Key Vault secret holding Service Pricipal credentials for gen1 lake
        gen2_sp_credentials_secret (str): The Key Vault secret holding Service Pricipal credentials for gen2 lake
        vault_name (str): The name of the vault from which to retrieve the secrets.
        timeout(int, optional): The amount of time (in seconds) to wait while running this task before
            a timeout occurs. Defaults to 3600.
    """

    def __init__(
        self,
        name: str,
        gen1_path: str,
        gen2_path: str,
        local_file_path: str = None,
        overwrite: bool = True,
        sep: str = "\t",
        gen1_sp_credentials_secret: str = None,
        gen2_sp_credentials_secret: str = None,
        vault_name: str = None,
        timeout: int = 3600,
        *args: List[any],
        **kwargs: Dict[str, Any]
    ):
        self.gen1_path = gen1_path
        self.local_file_path = local_file_path or self.slugify(name) + ".csv"
        self.gen2_path = gen2_path
        self.overwrite = overwrite
        self.sep = sep
        self.gen1_sp_credentials_secret = gen1_sp_credentials_secret
        self.gen2_sp_credentials_secret = gen2_sp_credentials_secret
        self.vault_name = vault_name
        self.timeout = timeout
        super().__init__(*args, name=name, **kwargs)
        self.gen_flow()

    @staticmethod
    def slugify(name):
        return name.replace(" ", "_").lower()

    def gen_flow(self) -> Flow:
        gen1_download_task = AzureDataLakeDownload(gen=1, timeout=self.timeout)
        gen1_download_task.bind(
            from_path=self.gen1_path,
            to_path=self.local_file_path,
            gen=1,
            sp_credentials_secret=self.gen1_sp_credentials_secret,
            vault_name=self.vault_name,
            flow=self,
        )
        add_ingestion_metadata.bind(path=self.local_file_path, sep=self.sep, flow=self)
        gen2_upload_task = AzureDataLakeUpload(gen=2, timeout=self.timeout)
        gen2_upload_task.bind(
            from_path=self.local_file_path,
            to_path=self.gen2_path,
            overwrite=self.overwrite,
            sp_credentials_secret=self.gen2_sp_credentials_secret,
            vault_name=self.vault_name,
            flow=self,
        )

        add_ingestion_metadata.set_upstream(gen1_download_task, flow=self)
        gen2_upload_task.set_upstream(add_ingestion_metadata, flow=self)



import logging

from viadot.flows import MultipleFlows


def test_multiple_flows_working(caplog):
    flow_list = [
        ["Flow of flows 1 test", "dev"],
        ["Flow of flows 2 - working", "dev"],
        ["Flow of flows 3", "dev"],
    ]
    flow = MultipleFlows(name="test", flows_list=flow_list)
    with caplog.at_level(logging.INFO):
        flow.run()
    assert "All of the tasks succeeded." in caplog.text


def test_multiple_flows_not_working(caplog):
    flow_list = [
        ["Flow of flows 1 test", "dev"],
        ["Flow of flows 2 test - not working", "dev"],
        ["Flow of flows 3", "dev"],
    ]
    flow = MultipleFlows(name="test", flows_list=flow_list)
    flow.run()
    assert "One of the flows has failed!" in caplog.text


def test_multiple_flows_working_one_flow(caplog):
    flow_list = [["Flow of flows 1 test", "dev"]]
    flow = MultipleFlows(name="test", flows_list=flow_list)
    with caplog.at_level(logging.INFO):
        flow.run()
    assert "All of the tasks succeeded." in caplog.text


def test_multiple_flows_last_not_working(caplog):
    flow_list = [
        ["Flow of flows 1 test", "dev"],
        ["Flow of flows 2 - working", "dev"],
        ["Flow of flows 2 test - not working", "dev"],
    ]
    flow = MultipleFlows(name="test", flows_list=flow_list)
    flow.run()
    assert "One of the flows has failed!" in caplog.text



import os
from pathlib import Path
from typing import Any, Dict, List

import pendulum
from prefect import Flow
from prefect.backend import set_key_value
from prefect.utilities import logging

from ..task_utils import (
    add_ingestion_metadata_task,
    df_get_data_types_task,
    df_map_mixed_dtypes_for_parquet,
    df_to_csv,
    df_to_parquet,
    dtypes_to_json_task,
    update_dtypes_dict,
)
from ..tasks import AzureDataLakeUpload, EurostatToDF

file_to_adls_task = AzureDataLakeUpload()
json_to_adls_task = AzureDataLakeUpload()


class EurostatToADLS(Flow):
    """Flow for downloading data from the Eurostat platform via HTTPS REST API (no credentials required)
    to a CSV or Parquet file. Then upload it to Azure Data Lake.
    """

    def __init__(
        self,
        name: str,
        dataset_code: str,
        params: dict = None,
        base_url: str = "https://ec.europa.eu/eurostat/api/dissemination/statistics/1.0/data/",
        requested_columns: list = None,
        output_file_extension: str = ".parquet",
        adls_dir_path: str = None,
        local_file_path: str = None,
        adls_file_name: str = None,
        adls_sp_credentials_secret: str = None,
        overwrite_adls: bool = False,
        if_exists: str = "replace",
        *args: List[Any],
        **kwargs: Dict[str, Any],
    ):
        """
        Args:
            name (str): The name of the flow.
            dataset_code(str): The code of eurostat dataset that has to be upload.
            params (Dict[str], optional):
                A dictionary with optional URL parameters. The key represents the parameter id, while the value is the code
                for a specific parameter, for example 'params = {'unit': 'EUR'}' where "unit" is the parameter that you would like to set
                and "EUR" is the code of the specific parameter. You can add more than one parameter, but only one code per parameter!
                So you CAN NOT provide list of codes as in example 'params = {'unit': ['EUR', 'USD', 'PLN']}'
                This parameter is REQUIRED in most cases to pull a specific dataset from the API.
                Both parameter and code has to provided as a string!
                Defaults to None.
            base_url (str): The base URL used to access the Eurostat API. This parameter specifies the root URL for all requests made to the API.
                It should not be modified unless the API changes its URL scheme.
                Defaults to "https://ec.europa.eu/eurostat/api/dissemination/statistics/1.0/data/"
            requested_columns (List[str], optional): List of columns that are needed from DataFrame - works as filter.
                The data are downloaded from Eurostat is the same structure every time. The filter is applied after the data is fetched.
            output_file_extension (str, optional): Output file extension - to allow selection of .csv for data
                    which is not easy to handle with parquet. Defaults to ".parquet".
            adls_dir_path (str, optional): Azure Data Lake destination folder/catalog path. Defaults to None.
            local_file_path (str, optional): Local destination path. Defaults to None.
            adls_file_name (str, optional): Name of file in ADLS. Defaults to None.
            adls_sp_credentials_secret (str, optional): The name of the Azure Key Vault secret containing a dictionary with
                ACCOUNT_NAME and Service Principal credentials (TENANT_ID, CLIENT_ID, CLIENT_SECRET) for the Azure Data Lake.
                Defaults to None.
            overwrite_adls (bool, optional): Whether to overwrite files in the lake. Defaults to False.
            if_exists (str, optional): What to do if the file exists. Defaults to "replace".
        """

        # EurostatToDF
        self.dataset_code = dataset_code
        self.params = params
        self.base_url = base_url
        self.requested_columns = requested_columns

        # AzureDataLakeUpload
        self.overwrite = overwrite_adls
        self.adls_sp_credentials_secret = adls_sp_credentials_secret
        self.if_exists = if_exists
        self.output_file_extension = output_file_extension
        self.now = str(pendulum.now("utc"))

        self.local_file_path = (
            local_file_path or self.slugify(name) + self.output_file_extension
        )
        self.local_json_path = self.slugify(name) + ".json"
        self.adls_dir_path = adls_dir_path

        if adls_file_name is not None:
            self.adls_file_path = os.path.join(adls_dir_path, adls_file_name)
            self.adls_schema_file_dir_file = os.path.join(
                adls_dir_path, "schema", Path(adls_file_name).stem + ".json"
            )
        else:
            self.adls_file_path = os.path.join(
                adls_dir_path, self.now + self.output_file_extension
            )
            self.adls_schema_file_dir_file = os.path.join(
                adls_dir_path, "schema", self.now + ".json"
            )

        super().__init__(name=name, *args, **kwargs)

        self.gen_flow()

    @staticmethod
    def slugify(name):
        return name.replace(" ", "_").lower()

    def gen_flow(self) -> Flow:
        df = EurostatToDF(
            dataset_code=self.dataset_code,
            params=self.params,
            base_url=self.base_url,
            requested_columns=self.requested_columns,
        )

        df = df.bind(flow=self)

        df_with_metadata = add_ingestion_metadata_task.bind(df, flow=self)
        dtypes_dict = df_get_data_types_task.bind(df_with_metadata, flow=self)

        if self.output_file_extension == ".parquet":
            df_to_be_loaded = df_map_mixed_dtypes_for_parquet(
                df_with_metadata, dtypes_dict, flow=self
            )
            df_to_file = df_to_parquet.bind(
                df=df_to_be_loaded,
                path=self.local_file_path,
                if_exists=self.if_exists,
                flow=self,
            )
        else:
            df_to_file = df_to_csv.bind(
                df=df_with_metadata,
                path=self.local_file_path,
                if_exists=self.if_exists,
                flow=self,
            )

        file_to_adls_task.bind(
            from_path=self.local_file_path,
            to_path=self.adls_file_path,
            overwrite=self.overwrite,
            sp_credentials_secret=self.adls_sp_credentials_secret,
            flow=self,
        )

        dtypes_updated = update_dtypes_dict(dtypes_dict, flow=self)
        dtypes_to_json_task.bind(
            dtypes_dict=dtypes_updated, local_json_path=self.local_json_path, flow=self
        )

        json_to_adls_task.bind(
            from_path=self.local_json_path,
            to_path=self.adls_schema_file_dir_file,
            overwrite=self.overwrite,
            sp_credentials_secret=self.adls_sp_credentials_secret,
            flow=self,
        )

        file_to_adls_task.set_upstream(df_to_file, flow=self)
        json_to_adls_task.set_upstream(dtypes_to_json_task, flow=self)
        set_key_value(key=self.adls_dir_path, value=self.adls_file_path)



from unittest import mock

from viadot.flows import ADLSGen1ToGen2


def test_adls_gen1_gen2_init(TEST_PARQUET_FILE_PATH_2):
    flow = ADLSGen1ToGen2(
        "test_adls_gen1_gen2_init",
        gen1_path=TEST_PARQUET_FILE_PATH_2,
        gen2_path=TEST_PARQUET_FILE_PATH_2,
    )
    assert flow


def test_adls_gen1_to_azure_sql_new_mock(
    TEST_PARQUET_FILE_PATH, TEST_PARQUET_FILE_PATH_2
):
    with mock.patch.object(ADLSGen1ToGen2, "run", return_value=True) as mock_method:
        instance = ADLSGen1ToGen2(
            "test_adls_gen1_gen2_init",
            gen1_path=TEST_PARQUET_FILE_PATH,
            gen2_path=TEST_PARQUET_FILE_PATH_2,
        )
        instance.run()
        mock_method.assert_called_with()



from datetime import datetime
from unittest import mock

import pytest

from viadot.tasks import GenesysToCSV


@pytest.fixture
def var_dictionary() -> None:
    """Function where variables are stored."""

    variables = {
        "start_date": datetime.now().strftime("%Y-%m-%d"),
        "end_date": datetime.now().strftime("%Y-%m-%d"),
        "v_list": [
            "xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxx1",
            "xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxx2",
            "xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxx3",
        ],
        "key_list": [
            "MainIntent",
            "SubIntent",
            "Final Sub Intent",
            "CustomerOutcomeTrack",
            "LastUtterance",
            "Final Main Intent",
        ],
        "post_data_list": [
            {
                "name": "AGENT_STATUS_DETAIL_VIEW",
                "timeZone": "UTC",
                "exportFormat": "CSV",
                "interval": "2023-03-01T00:00:00/2023-03-02T00:00:00",
                "period": "PT30M",
                "viewType": "AGENT_STATUS_DETAIL_VIEW",
                "filter": {"userIds": ["9eb0fe4e-937e-4443-a5a4-d1b5dbd76520"]},
                "read": True,
                "locale": "en-us",
                "hasFormatDurations": False,
                "hasSplitFilters": True,
                "excludeEmptyRows": True,
                "hasSummaryRow": False,
                "csvDelimiter": "COMMA",
                "hasCustomParticipantAttributes": True,
            }
        ],
        "post_data_list_2": [
            {
                "interval": "2023-03-09T00:00:00/2023-03-10T00:00:00",
                "paging": {"pageSize": 100, "pageNumber": 1},
            }
        ],
    }
    return variables


class MockGenesysTask:
    report_data = ([[None, "COMPLETED"], [None, "COMPLETED"]],)

    def genesys_api_connection(post_data_list, end_point, method="POST"):
        if method == "GET":
            report = {
                "id": "xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx",
                "startTime": "2023-06-28T10:59:48.194Z",
                "participants": [
                    {
                        "id": "xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx",
                        "startTime": "2023-06-28T10:59:48.194Z",
                        "connectedTime": "2023-06-28T10:59:48.194Z",
                        "externalContactId": "xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx",
                        "queueId": "xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx",
                        "queueName": "dev_name",
                        "purpose": "customer",
                        "wrapupRequired": None,
                        "mediaRoles": ["full"],
                        "attributes": {
                            "MainIntent3": "mainintent3",
                            "SubIntent3": "subintent3",
                            "SubIntent2": "subintent2",
                            "MainIntent4": "mainintent4",
                            "MainIntent1": "mainintent1",
                            "SubIntent1": "subintent1",
                            "MainIntent2": "mainintent2",
                            "Final Sub Intent": "finalsubintent",
                            "SubIntent4": "subintent4",
                            "CustomerOutcomeTrack4": "customeroutcome4",
                            "CustomerOutcomeTrack3": "customeroutcome3",
                            "CustomerOutcomeTrack2": "customeroutcome2",
                            "CustomerOutcomeTrack1": "customeroutcome1",
                            "LastUtterance2": "lastutterance2",
                            "reached": "reach",
                            "LastUtterance1": "lastutterance1",
                            "name": "dev_name",
                            "Final Main Intent": "finalmainintent",
                            "LastUtterance4": "lastutterance4",
                            "LastUtterance3": "lastutterance3",
                            "LOB": "lob",
                            "memberId": "123456789",
                        },
                        "calls": [],
                        "callbacks": [],
                        "chats": [],
                        "cobrowsesessions": [],
                        "emails": [],
                        "messages": [],
                    }
                ],
            }
        else:
            report = {
                "conversations": [
                    {
                        "conversationEnd": "2020-01-01T00:00:00.00Z",
                        "conversationId": "xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx",
                        "conversationStart": "2020-01-01T00:00:00.00Z",
                        "divisionIds": [
                            "xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx",
                            "xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx",
                        ],
                        "mediaStatsMinConversationMos": 4.379712366260067,
                        "mediaStatsMinConversationRFactor": 79.03050231933594,
                        "originatingDirection": "inbound",
                        "participants": [
                            {
                                "externalContactId": "xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx",
                                "participantId": "xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx",
                                "participantName": "Mobile Number, Country",
                                "purpose": "customer",
                                "sessions": [
                                    {
                                        "agentBullseyeRing": 1,
                                        "ani": "tel:+xxxxxxxxxxx",
                                        "direction": "inbound",
                                        "dnis": "tel:+xxxxxxxxxxx",
                                        "edgeId": "xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx",
                                        "mediaType": "voice",
                                        "protocolCallId": "xxxxxxxxxxxxxxxxxxx@xx.xxx.xxx.xxx",
                                        "provider": "Edge",
                                        "remoteNameDisplayable": "Mobile Number, Country",
                                        "requestedRoutings": ["Standard"],
                                        "routingRing": 1,
                                        "selectedAgentId": "xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx",
                                        "sessionDnis": "tel:+xxxxxxxxxxx",
                                        "sessionId": "xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx",
                                        "usedRouting": "Standard",
                                        "mediaEndpointStats": [
                                            {
                                                "codecs": ["audio/opus"],
                                                "eventTime": "2020-01-01T00:00:00.00Z",
                                                "maxLatencyMs": 30,
                                                "minMos": 4.882504366160681,
                                                "minRFactor": 92.44775390625,
                                                "receivedPackets": 229,
                                            },
                                        ],
                                        "metrics": [
                                            {
                                                "emitDate": "2020-01-01T00:00:00.00Z",
                                                "name": "nConnected",
                                                "value": 1,
                                            },
                                        ],
                                        "segments": [
                                            {
                                                "conference": False,
                                                "segmentEnd": "2020-01-01T00:00:00.00Z",
                                                "segmentStart": "2020-01-01T00:00:00.00Z",
                                                "segmentType": "system",
                                            },
                                            {
                                                "conference": False,
                                                "disconnectType": "peer",
                                                "queueId": "xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx",
                                                "segmentEnd": "2020-01-01T00:00:00.00Z",
                                                "segmentStart": "2020-01-01T00:00:00.00Z",
                                                "segmentType": "interact",
                                            },
                                        ],
                                    }
                                ],
                            },
                            {
                                "participantId": "xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx",
                                "participantName": "xxxxxxxxxxxxxxxxxxxxx",
                                "purpose": "ivr",
                                "sessions": [
                                    {
                                        "ani": "tel:+xxxxxxxxxxx",
                                        "direction": "inbound",
                                        "dnis": "tel:+xxxxxxxxxxx",
                                        "edgeId": "xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx",
                                        "mediaType": "voice",
                                        "peerId": "xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx",
                                        "protocolCallId": "xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx",
                                        "provider": "Edge",
                                        "remote": "Mobile Number, Country",
                                        "remoteNameDisplayable": "xxxxxxxx, Country",
                                        "sessionDnis": "tel:+xxxxxxxxxxx",
                                        "sessionId": "xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx",
                                        "mediaEndpointStats": [
                                            {
                                                "codecs": ["audio/opus"],
                                                "eventTime": "2020-01-01T00:00:00.00Z",
                                                "maxLatencyMs": 30,
                                                "minMos": 4.429814389713434,
                                                "minRFactor": 79.03050231933594,
                                                "receivedPackets": 229,
                                            }
                                        ],
                                        "flow": {
                                            "endingLanguage": "lt-lt",
                                            "entryReason": "tel:+xxxxxxxxxxx",
                                            "entryType": "dnis",
                                            "exitReason": "TRANSFER",
                                            "flowId": "xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx",
                                            "flowName": "xxxxxxxxxxxxxxxxxxxxx",
                                            "flowType": "INBOUNDCALL",
                                            "flowVersion": "22.0",
                                            "startingLanguage": "en-us",
                                            "transferTargetAddress": "xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx",
                                            "transferTargetName": "xxxxxxxxxxxxxxxxxxxxx",
                                            "transferType": "ACD",
                                        },
                                        "metrics": [
                                            {
                                                "emitDate": "2020-01-01T00:00:00.00Z",
                                                "name": "nFlow",
                                                "value": 1,
                                            },
                                        ],
                                        "segments": [
                                            {
                                                "conference": False,
                                                "segmentEnd": "2020-01-01T00:00:00.00Z",
                                                "segmentStart": "2020-01-01T00:00:00.00Z",
                                                "segmentType": "system",
                                            },
                                        ],
                                    }
                                ],
                            },
                        ],
                    }
                ],
                "totalHits": 100,
            }
        return report

    def get_reporting_exports_data():
        pass

    def delete_all_reporting_exports():
        pass

    def download_all_reporting_exports(path):
        return [
            "V_D_PROD_FB_QUEUE_CALLBACK.csv",
            "V_D_PROD_FB_QUEUE_CHAT.csv",
        ]

    def delete_all_reporting_exports():
        pass


@mock.patch.object(GenesysToCSV, "run")
@pytest.mark.to_csv
def test_genesys_to_csv(mock_reuturn):
    mock_reuturn.return_value = [
        "V_D_PROD_FB_QUEUE_CALLBACK.csv",
        "V_D_PROD_FB_QUEUE_CHAT.csv",
    ]
    genesys_to_csv = GenesysToCSV()
    files_name_list = genesys_to_csv.run()
    assert isinstance(files_name_list, list)


@mock.patch("viadot.tasks.genesys.Genesys", return_value=MockGenesysTask)
@pytest.mark.files
def test_genesys_files_type(mock_genesys, var_dictionary):
    to_csv = GenesysToCSV()
    file_name = to_csv.run(
        view_type="agent_status_detail_view",
        view_type_time_sleep=5,
        post_data_list=var_dictionary["post_data_list"],
        start_date=var_dictionary["start_date"],
    )
    mock_genesys.assert_called_once()
    assert len(file_name) > 1


@mock.patch("viadot.tasks.genesys.Genesys", return_value=MockGenesysTask)
@pytest.mark.conv
def test_genesys_conversations(mock_genesys, var_dictionary):
    to_csv = GenesysToCSV()
    file_name = to_csv.run(
        view_type=None,
        end_point="analytics/conversations/details/query",
        post_data_list=var_dictionary["post_data_list_2"],
        start_date=var_dictionary["start_date"],
    )
    date = var_dictionary["start_date"].replace("-", "")

    mock_genesys.assert_called_once()
    assert file_name[0] == f"conversations_detail_{date}".upper() + ".csv"


@mock.patch("viadot.tasks.genesys.Genesys", return_value=MockGenesysTask)
@pytest.mark.conv
def test_genesys_webmsg(mock_genesys, var_dictionary):
    to_csv = GenesysToCSV()
    file_name = to_csv.run(
        view_type=None,
        end_point="conversations",
        conversationId_list=var_dictionary["v_list"],
        post_data_list=[""],
        key_list=var_dictionary["key_list"],
        start_date=var_dictionary["start_date"],
        end_date=var_dictionary["end_date"],
    )

    start = var_dictionary["start_date"].replace("-", "")
    end = var_dictionary["end_date"].replace("-", "")

    mock_genesys.assert_called_once()
    assert file_name[0] == f"WEBMESSAGE_{start}-{end}.csv"



import os
from unittest import mock

import pandas as pd
import pytest

from viadot.flows import MediatoolToADLS

DATA = {"country": ["DK", "DE"], "sales": [3, 4]}
ADLS_FILE_NAME = "test_mediatool.parquet"
ADLS_DIR_PATH = "raw/tests/"


@mock.patch(
    "viadot.tasks.MediatoolToDF.run",
    return_value=pd.DataFrame(data=DATA),
)
@pytest.mark.run
def test_mediatool_to_adls_run_flow(mocked_class):
    flow = MediatoolToADLS(
        "test_mediatool_to_adls_flow_run",
        organization_ids=["1000000001", "200000001"],
        media_entries_columns=["id", "name", "num"],
        mediatool_credentials_key="MEDIATOOL-TESTS",
        overwrite_adls=True,
        adls_dir_path=ADLS_DIR_PATH,
        adls_file_name=ADLS_FILE_NAME,
    )
    result = flow.run()
    assert result.is_successful()
    os.remove("test_mediatool_to_adls_flow_run.parquet")
    os.remove("test_mediatool_to_adls_flow_run.json")



from unittest.mock import Mock, patch

import pandas as pd
import pytest

from viadot.sources import BusinessCore


@pytest.fixture(scope="module")
def business_core():
    return BusinessCore(
        url="https://api.businesscore.ae/api/GetCustomerData",
        filters_dict={
            "BucketCount": 10,
            "BucketNo": 1,
            "FromDate": None,
            "ToDate": None,
        },
        credentials={"username": "test", "password": "test123"},
    )


@patch("viadot.sources.business_core.handle_api_response")
def test_generate_token(mock_api_response, business_core):
    mock_api_response.return_value = Mock(text='{"access_token": "12345"}')
    token = business_core.generate_token()
    assert token == "12345"


def test_clean_filters_dict(business_core):
    filters = business_core.clean_filters_dict()
    assert filters == {
        "BucketCount": 10,
        "BucketNo": 1,
        "FromDate": "&",
        "ToDate": "&",
    }


def test_to_df(business_core):
    with patch.object(
        business_core,
        "get_data",
        return_value={"MasterDataList": [{"id": 1, "name": "John Doe"}]},
    ):
        df = business_core.to_df()
        assert isinstance(df, pd.DataFrame)
        assert len(df.columns) == 2
        assert len(df) == 1
        assert df["id"].tolist() == [1]
        assert df["name"].tolist() == ["John Doe"]



import os

from prefect.tasks.secrets import PrefectSecret

from viadot.flows import SalesforceToADLS
from viadot.tasks import AzureDataLakeRemove

ADLS_FILE_NAME = "test_salesforce.parquet"
ADLS_DIR_PATH = "raw/tests/"


def test_salesforce_to_adls():
    credentials_secret = PrefectSecret(
        "AZURE_DEFAULT_ADLS_SERVICE_PRINCIPAL_SECRET"
    ).run()

    flow = SalesforceToADLS(
        "test_salesforce_to_adls_run_flow",
        query="SELECT IsDeleted, FiscalYear FROM Opportunity LIMIT 50",
        adls_sp_credentials_secret=credentials_secret,
        adls_dir_path=ADLS_DIR_PATH,
        adls_file_name=ADLS_FILE_NAME,
    )

    result = flow.run()
    assert result.is_successful()

    os.remove("test_salesforce_to_adls_run_flow.parquet")
    os.remove("test_salesforce_to_adls_run_flow.json")
    rm = AzureDataLakeRemove(
        path=ADLS_DIR_PATH + ADLS_FILE_NAME,
        vault_name="azuwevelcrkeyv001s",
    )
    rm.run(sp_credentials_secret=credentials_secret)



from typing import Any, Dict, List, Literal

from prefect import Flow

from viadot.flows.adls_to_azure_sql import df_to_csv_task
from viadot.task_utils import add_ingestion_metadata_task
from viadot.tasks import AzureDataLakeUpload, AzureSQLCreateTable, BCPTask, SftpToDF


class SftpToAzureSQL(Flow):
    def __init__(
        self,
        name: str,
        from_path: str = None,
        file_name: str = None,
        columns: List[str] = None,
        sep: str = "\t",
        remove_tab: bool = True,
        dtypes: Dict[str, Any] = None,
        table: str = None,
        schema: str = None,
        if_exists: Literal["fail", "replace", "append", "delete"] = "fail",
        sftp_credentials_secret: Dict[str, Any] = None,
        sftp_credentials: Dict[str, Any] = None,
        sqldb_credentials_secret: str = None,
        on_bcp_error: Literal["skip", "fail"] = "fail",
        error_log_file_path: str = "SFTP_logs.log",
        vault_name: str = None,
        timeout: int = 3600,
        *args,
        **kwargs,
    ):
        """
        Bulk insert data from SFTP server into an Azure SQL Database table.
        This task also creates the table if it doesn't exist.

        Args:
            name (str): The name of the flow.
            from_path (str, optional): Path to the file in SFTP server. Defaults to None.
            file_name (str, optional): File name for local file. Defaults to None.
            columns (List[str], optional): Columns to read from the file. Defaults to None.
            sep (str, optional): The separator to use to read the CSV file. Defaults to "\t".
            remove_tab (bool, optional): Whether to remove tab delimiters from the data. Defaults to False.
            dtypes (dict, optional): Which custom data types should be used for SQL table creation task.
            table (str, optional): Destination table. Defaults to None.
            schema (str, optional): Destination schema. Defaults to None.
            if_exists (Literal, optional): What to do if the table already exists. Defaults to "replace".
            sftp_credentials_secret (str, optional): The name of the Azure Key Vault secret containing a dictionary credentials for SFTP connection. Defaults to None.
            sftp_credentials (Dict[str, Any], optional): SFTP server credentials. Defaults to None.
            sqldb_credentials_secret (str, optional): The name of the Azure Key Vault secret containing a dictionary with
            on_bcp_error (Literal["skip", "fail"], optional): What to do if error occurs. Defaults to "fail".
            error_log_file_path (string, optional): Full path of an error file. Defaults to "./log_file.log".
            vault_name (str, optional): The name of the vault from which to obtain the secret. Defaults to None.
            timeout(int, optional): The amount of time (in seconds) to wait while running this task before
                a timeout occurs. Defaults to 3600.
        """
        # SFTP
        self.from_path = from_path
        self.sftp_credentials_secret = sftp_credentials_secret
        self.sftp_credentials = sftp_credentials
        self.columns = columns
        # File args
        if file_name is None:
            self.file_name = from_path.split("/")[-1]

        else:
            self.file_name = file_name

        self.sep = sep
        self.remove_tab = remove_tab
        self.timeout = timeout

        # Read schema
        self.schema = schema
        self.table = table
        self.dtypes = dtypes

        # AzureSQLCreateTable
        self.table = table
        self.schema = schema
        self.if_exists = if_exists
        self.vault_name = vault_name

        # BCPTask
        self.sqldb_credentials_secret = sqldb_credentials_secret
        self.on_bcp_error = on_bcp_error
        self.error_log_file_path = error_log_file_path

        super().__init__(
            name=name,
            *args,
            **kwargs,
        )

        self.gen_flow()

    @staticmethod
    def _map_if_exists(if_exists: str) -> str:
        mapping = {"append": "skip"}
        return mapping.get(if_exists, if_exists)

    def __call__(self, *args, **kwargs):
        """Download file-like object from SFTP server and load data into Azure SQL database."""
        return super().__call__(*args, **kwargs)

    def gen_flow(self) -> Flow:
        sftp = SftpToDF(
            sftp_credentials_secret=self.sftp_credentials_secret,
            credentials=self.sftp_credentials,
            timeout=self.timeout,
        )
        df = sftp.bind(
            from_path=self.from_path,
            columns=self.columns,
            flow=self,
        )
        df_with_metadata = add_ingestion_metadata_task.bind(df, flow=self)
        df_to_csv_task.bind(
            df=df_with_metadata,
            remove_tab=self.remove_tab,
            path=self.file_name,
            flow=self,
        )

        create_table_task = AzureSQLCreateTable(timeout=self.timeout)
        create_table_task.bind(
            schema=self.schema,
            table=self.table,
            dtypes=self.dtypes,
            if_exists=self._map_if_exists(self.if_exists),
            credentials_secret=self.sqldb_credentials_secret,
            vault_name=self.vault_name,
            flow=self,
        )

        bulk_insert_task = BCPTask(timeout=self.timeout)
        bulk_insert_task.bind(
            path=self.file_name,
            schema=self.schema,
            table=self.table,
            error_log_file_path=self.error_log_file_path,
            on_error=self.on_bcp_error,
            credentials_secret=self.sqldb_credentials_secret,
            flow=self,
        )

        df_with_metadata.set_upstream(df, flow=self)
        df_to_csv_task.set_upstream(df_with_metadata, flow=self)
        create_table_task.set_upstream(df_to_csv_task, flow=self)
        bulk_insert_task.set_upstream(create_table_task, flow=self)


class SftpToADLS(Flow):
    def __init__(
        self,
        name: str,
        from_path: str = None,
        file_name: str = None,
        sep: str = "\t",
        remove_tab: bool = True,
        overwrite: bool = True,
        to_path: str = None,
        columns: List[str] = None,
        sftp_credentials_secret: Dict[str, Any] = None,
        sftp_credentials: Dict[str, Any] = None,
        sp_credentials_secret: str = None,
        vault_name: str = None,
        timeout: int = 3600,
        *args,
        **kwargs,
    ):
        """
        Bulk insert data from SFTP server into an Azure SQL Database table.
        This task also creates the table if it doesn't exist.

        Args:
            name (str): The name of the flow.
            from_path (str, optional): Path to the file in SFTP server. Defaults to None.
            file_name (str, optional): File name for local file. Defaults to None.
            sep (str, optional): The separator to use to read the CSV file. Defaults to "\t".
            remove_tab (bool, optional): Whether to remove tab delimiters from the data. Defaults to False.
            overwrite (bool, optional): Whether to overwrite files in the lake. Defaults to False.
            to_path (str, optional): The destination path in ADLS. Defaults to None.
            columns (List[str], optional): Columns to read from the file. Defaults to None.
            sftp_credentials_secret (str, optional): The name of the Azure Key Vault secret containing a dictionary credentials for SFTP connection. Defaults to None.
            sftp_credentials (Dict[str, Any], optional): SFTP server credentials. Defaults to None.
            sp_credentials_secret (str, optional): The name of the Azure Key Vault secret containing a dictionary.
            vault_name (str, optional): The name of the vault from which to obtain the secret. Defaults to None.
            timeout(int, optional): The amount of time (in seconds) to wait while running this task before
                a timeout occurs. Defaults to 3600.
        """
        # SFTP
        self.from_path = from_path
        self.sftp_credentials_secret = sftp_credentials_secret
        self.sftp_credentials = sftp_credentials
        self.columns = columns
        self.timeout = timeout

        # File args
        if file_name is None:
            self.file_name = from_path.split("/")[-1]

        else:
            self.file_name = file_name

        self.sep = sep
        self.remove_tab = remove_tab

        # ADLS
        self.to_path = to_path
        self.sp_credentials_secret = sp_credentials_secret
        self.vault_name = vault_name
        self.overwrite = overwrite

        super().__init__(
            name=name,
            *args,
            **kwargs,
        )

        self.gen_flow()

    def __call__(self, *args, **kwargs):
        """Download file-like object from SFTP server and load data into ADLS."""
        return super().__call__(*args, **kwargs)

    def gen_flow(self) -> Flow:
        ftp = SftpToDF(
            sftp_credentials_secret=self.sftp_credentials_secret,
            credentials=self.sftp_credentials,
            timeout=self.timeout,
        )
        df = ftp.bind(
            from_path=self.from_path,
            columns=self.columns,
            flow=self,
        )
        df_to_csv_task.bind(
            df=df, remove_tab=self.remove_tab, path=self.file_name, flow=self
        )

        upload_to_adls = AzureDataLakeUpload(timeout=self.timeout)
        upload_df = upload_to_adls.bind(
            from_path=self.file_name,
            to_path=self.to_path,
            sp_credentials_secret=self.sp_credentials_secret,
            overwrite=self.overwrite,
            vault_name=self.vault_name,
            flow=self,
        )

        df_to_csv_task.set_upstream(df, flow=self)
        upload_df.set_upstream(df_to_csv_task, flow=self)



from typing import Any, Dict, List, Literal

from prefect import Flow, task, unmapped

from viadot.task_utils import concat_dfs, df_to_csv, df_to_parquet, set_new_kv
from viadot.tasks import AzureDataLakeUpload, SAPRFCToDF


class SAPRFCToADLS(Flow):
    def __init__(
        self,
        name: str,
        query: str = None,
        rfc_sep: str = None,
        rfc_replacement: str = "-",
        func: str = "RFC_READ_TABLE",
        rfc_total_col_width_character_limit: int = 400,
        rfc_unique_id: List[str] = None,
        sap_credentials: dict = None,
        output_file_extension: str = ".parquet",
        local_file_path: str = None,
        file_sep: str = "\t",
        if_exists: Literal["append", "replace", "skip"] = "replace",
        adls_path: str = None,
        overwrite: bool = False,
        adls_sp_credentials_secret: str = None,
        vault_name: str = None,
        update_kv: bool = False,
        filter_column: str = None,
        timeout: int = 3600,
        alternative_version: bool = False,
        *args: List[any],
        **kwargs: Dict[str, Any],
    ):
        """
        Flow for downloading data from SAP database using the RFC protocol and uploading it to Azure Data Lake.

        Note that only a very limited subset of SQL is supported:
            - aliases
            - where clauses combined using the AND operator
            - limit & offset

        Unsupported:
            - aggregations
            - joins
            - subqueries
            - etc.

        Args:
            name (str): The name of the flow.
            query (str): Query to be executed with pyRFC. Defaults to None.
            rfc_sep(str, optional): Which separator to use when querying SAP. If not provided, multiple options are automatically tried.
            rfc_replacement (str, optional): In case of sep is on a columns, set up a new character to replace
                inside the string to avoid flow breakdowns. Defaults to "-".
            func (str, optional): SAP RFC function to use. Defaults to "RFC_READ_TABLE".
            rfc_total_col_width_character_limit (int, optional): Number of characters by which query will be split in chunks in case of too many columns
                for RFC function. According to SAP documentation, the limit is 512 characters. However, we observed SAP raising an exception
                even on a slightly lower number of characters, so we add a safety margin. Defaults to 400.
            rfc_unique_id  (List[str], optional): Reference columns to merge chunks Data Frames. These columns must to be unique. If no columns are provided
                in this parameter, all data frame columns will by concatenated. Defaults to None.
                Example:
                --------
                SAPRFCToADLS(
                    ...
                    rfc_unique_id=["VBELN", "LPRIO"],
                    ...
                    )
            sap_credentials (dict, optional): The credentials to use to authenticate with SAP. By default, they're taken from the local viadot config.
            output_file_extension (str, optional): Output file extension - to allow selection of .csv for data which is not easy to handle with parquet. Defaults to ".parquet".
            local_file_path (str, optional): Local destination path. Defaults to None.
            file_sep(str, optional): The separator to use in the CSV. Defaults to "\t".
            if_exists (Literal["append", "replace", "skip"], optional): What to do if the table exists. Defaults to "replace".
            adls_path(str, optional): Azure Data Lake destination file path. Defaults to None.
            overwrite(bool, optional) Whether to overwrite the file in ADLS. Defaults to False.
            adls_sp_credentials_secret(str, optional): The name of the Azure Key Vault secret containing a dictionary with ACCOUNT_NAME and Service Principal
            credentials (TENANT_ID, CLIENT_ID, CLIENT_SECRET) for the Azure Data Lake.Defaults to None.
            vault_name(str, optional): The name of the vault from which to obtain the secrets. Defaults to None.
            update_kv (bool, optional): Whether or not to update key value on Prefect. Defaults to False.
            filter_column (str, optional): Name of the field based on which key value will be updated. Defaults to None.
            timeout(int, optional): The amount of time (in seconds) to wait while running this task before
                a timeout occurs. Defaults to 3600.
            alternative_version (bool, optional): Enable the use version 2 in source. Defaults to False.
        """
        self.query = query
        self.rfc_sep = rfc_sep
        self.rfc_replacement = rfc_replacement
        self.func = func
        self.rfc_total_col_width_character_limit = rfc_total_col_width_character_limit
        self.rfc_unique_id = rfc_unique_id
        self.sap_credentials = sap_credentials
        self.output_file_extension = output_file_extension
        self.local_file_path = local_file_path
        self.file_sep = file_sep
        self.if_exists = if_exists
        self.adls_path = adls_path
        self.overwrite = overwrite
        self.adls_sp_credentials_secret = adls_sp_credentials_secret
        self.vault_name = vault_name
        self.timeout = timeout
        self.alternative_version = alternative_version

        self.update_kv = update_kv
        self.filter_column = filter_column

        super().__init__(*args, name=name, **kwargs)

        self.gen_flow()

    def gen_flow(self) -> Flow:
        download_sap_task = SAPRFCToDF(timeout=self.timeout)
        df = download_sap_task(
            query=self.query,
            sep=self.rfc_sep,
            replacement=self.rfc_replacement,
            func=self.func,
            rfc_total_col_width_character_limit=self.rfc_total_col_width_character_limit,
            rfc_unique_id=self.rfc_unique_id,
            alternative_version=self.alternative_version,
            credentials=self.sap_credentials,
            flow=self,
        )

        if self.output_file_extension == ".parquet":
            df_to_file = df_to_parquet.bind(
                df=df,
                path=self.local_file_path,
                if_exists=self.if_exists,
                flow=self,
            )
        else:
            df_to_file = df_to_csv.bind(
                df=df,
                sep=self.file_sep,
                path=self.local_file_path,
                if_exists=self.if_exists,
                flow=self,
            )

        file_to_adls_task = AzureDataLakeUpload(timeout=self.timeout)
        adls_upload = file_to_adls_task.bind(
            from_path=self.local_file_path,
            to_path=self.adls_path,
            overwrite=self.overwrite,
            sp_credentials_secret=self.adls_sp_credentials_secret,
            flow=self,
        )

        df_to_file.set_upstream(df, flow=self)
        adls_upload.set_upstream(df_to_file, flow=self)

        if self.update_kv == True:
            set_new_kv.bind(
                kv_name=self.name,
                df=df,
                filter_column=self.filter_column,
                flow=self,
            )
            set_new_kv.set_upstream(adls_upload, flow=self)



import xml.etree.ElementTree as ET
from typing import Any, Dict, List, Optional
from xml.etree.ElementTree import fromstring

import pandas as pd
from prefect import Task
from prefect.utilities.tasks import defaults_from_attrs

from ..sources import Epicor


class EpicorOrdersToDF(Task):
    def __init__(
        self,
        base_url: str,
        filters_xml: str,
        credentials: Dict[str, Any] = None,
        config_key: str = None,
        validate_date_filter: bool = True,
        start_date_field: str = "BegInvoiceDate",
        end_date_field: str = "EndInvoiceDate",
        timeout: int = 3600,
        *args,
        **kwargs,
    ) -> pd.DataFrame:
        """
        Task for downloading and parsing orders data from Epicor API to a pandas DataFrame.

        Args:
            name (str): The name of the flow.
            base_url (str, required): Base url to Epicor Orders.
            filters_xml (str, required): Filters in form of XML. The date filter is necessary.
            credentials (Dict[str, Any], optional): Credentials to connect with Epicor Api containing host, port, username and password. Defaults to None.
            config_key (str, optional): Credential key to dictionary where details are stored. Defauls to None.
            validate_date_filter (bool, optional): Whether or not validate xml date filters. Defaults to True.
            start_date_field (str, optional) The name of filters filed containing start date. Defaults to "BegInvoiceDate".
            end_date_field (str, optional) The name of filters filed containing end date. Defaults to "EndInvoiceDate".
            timeout(int, optional): The amount of time (in seconds) to wait while running this task before
                a timeout occurs. Defaults to 3600.

        Returns:
            pd.DataFrame: DataFrame with parsed API output
        """
        self.credentials = credentials
        self.config_key = config_key
        self.base_url = base_url
        self.filters_xml = filters_xml
        self.validate_date_filter = validate_date_filter
        self.start_date_field = start_date_field
        self.end_date_field = end_date_field
        super().__init__(
            name="epicor_orders_to_df",
            timeout=timeout,
            *args,
            **kwargs,
        )

    def __call__(self, *args, **kwargs):
        """Load Epicor Orders to DF"""
        return super().__call__(*args, **kwargs)

    @defaults_from_attrs(
        "credentials",
        "config_key",
        "base_url",
        "filters_xml",
        "validate_date_filter",
        "start_date_field",
        "end_date_field",
    )
    def run(
        self,
        credentials: Dict[str, Any] = None,
        config_key: str = None,
        base_url: str = None,
        filters_xml: str = None,
        validate_date_filter: bool = True,
        start_date_field: str = None,
        end_date_field: str = None,
    ):
        epicor = Epicor(
            credentials=credentials,
            config_key=config_key,
            base_url=base_url,
            filters_xml=filters_xml,
            validate_date_filter=validate_date_filter,
            start_date_field=start_date_field,
            end_date_field=end_date_field,
        )
        return epicor.to_df()
